//#include <QtGui>

#include <cmath> // log()
#include <QDebug>
#include <algorithm>
#include <QApplication> // отправить события

//using namespace std;
#include "nr.h"
#include "spreadsheet.h"
#include "cluster.h"
#include "seldialog.h"

// внешние переменные --------------
extern QApplication *gpApplication;// обработка событий
extern int gStopCalculation; // внешний флаг Останов
extern int gProgressCalc; // прогресс (кол-во столбцов)
// останов вычислений
int StopCalculation();

//_____________________________________
// конструктор выполняет все вычисления
// ВХОД:
//  sp - таблица
//  sc - выделенные ячейки в таблице
// ВЫХОД:
//  err - код ошибки (успех=0)
//  res - таблица результатов в HTML-формате
claster::claster(Spreadsheet *sp, selcel *sc, int &err, QString &res){
    err=0;
   spreadsheet = sp;
   scell = sc;
   if(scell->rows==2){// горизонтальная таблица
       NumFactors = scell->cols;// кол-во столбцов выделенной области
       // rowSum[0] - сумма выборки контроль, [1] - случай
       sum1 = scell->rowSum[0];
        sum2 = scell->rowSum[1];

   }else if(scell->cols==2){// вертикальная таблица
       NumFactors = scell->rows;// кол-во столбцов выделенной области
       // colSum[0] - сумма выборки контроль, [1] - случай
       sum1 = scell->colSum[0];
        sum2 = scell->colSum[1];
   }else{// ошибка
       err=1;
       res = QObject::tr("<br />Selection error");
       return;
   }
   if(scell->leftcol==0){// отсутствуют названия выборок
       col0 = 1;
   }else{// в первом столбце - названия выборок
       col0 = 0;
   }
   R1 = scell->R1;// значения выборки контроль
   R2 = scell->R2;// ... - случай
   log0 = -lnLambda(sum1, sum2);

   clsInd = new int [NumFactors * NumFactors];
   pval = new double [NumFactors+1];
   pUnit1 = new fUnit [NumFactors];// 0 в выборке контроль
   pUnit2 = new fUnit [NumFactors];// нулей нет
   pUnit3 = new fUnit [NumFactors];// 0 в выборке случай
   fnum1 = fnum2 = fnum3 = 0;// размеры pUnitN
   // цикл заполнения контейнеров pUnitN
   // NumFactors - кол-во факторов
   for(int i=0; i<NumFactors; i++){
       for(int j=0; j<NumFactors; j++){
           clsInd[j]=0;
       }
      fUnit *pf;
      if(R1[i]==0){
         pf = &pUnit1[fnum1++];
      }else if(R2[i]==0){
         pf = &pUnit3[fnum3++];
      }else{
         pf = &pUnit2[fnum2++];
      }
      pf->ind = i;
      pf->r1 = R1[i];
      pf->r2 = R2[i];
   }

   sortClaster();  // заполнение списка кластеров
   err = copyFactor();// основные вычисления
   res = tblres;// возвращение результата
}
//_________________________________
// деструктор освобождает память new
claster::~claster(){
    delete [] clsInd;
   delete [] pval;
   delete [] pUnit1;
   delete [] pUnit2;
   delete [] pUnit3;
}
//===================================================
// ОСНОВНОЙ цикл вычислений
// вместо копирования в таблицу копировать в контейнер
// res - результат
int claster::copyFactor(){
    int numClusters;     // кол-во кластеров (>=2)
   double minPval;  // min(pval)
    // цикл объединения строк
   for(int numc=NumFactors; numc>1; numc--){
      // numc - шаг алгоритма (кол-во кластеров)
      if(StopCalculation()){// внешний останов
         return 1;
      }
      merge(numc);// слияние 2-х ближайших кластеров
      // вычисление значимости набора кластеров
      pval[numc] = znach(numc);
      if(numc == NumFactors){// первый шаг
         numClusters = numc;
         minPval = pval[numc];
      }else{
         if(pval[numc]<minPval){
            minPval = pval[numc];
            numClusters = numc;
         }
      }
      // вывод одной строки с номерами кластеров
      outres(numc, numClusters);
      gProgressCalc = NumFactors-numc+1;   // прогресс вычислений
   }
   setRes(numClusters);// вывод принадлежности факторов кластерам
   OddsRatio(numClusters);// расчет отношения шансов

   return 0;// успех
}
//________________________________
// класс сравнения в 1-ой строке нули
class order1{
public: bool operator()(const fUnit &p, const fUnit &q) const
   {
      return p.r2 > q.r2;
   }
};
//________________________________
// класс сравнения, нулей нет
class order2{
public: bool operator()(const fUnit &p, const fUnit &q) const
   {
      return (double) p.r1/p.r2 < (double) q.r1/q.r2;
   }
};
//________________________________
// класс сравнения в 2-ой строке нули
class order3{
public: bool operator()(const fUnit &p, const fUnit &q) const
   {
      return p.r1 < q.r1;
   }
};
//________________________________
// возврат хи-квадрат расстояния между кластерами it и it+1
//
double claster::chi2dist(list<fUnit>::iterator it){
   double n0, n1, m0, m1, C1, C2, R1, R2, N,
         n0e, n1e, m0e, m1e, chi;

   n0 = (*it).r1;   n1 = (*it).r2;   C1 = n0 + n1;

   it++;

   m0 = (*it).r1;   m1 = (*it).r2;   C2 = m0 + m1;

   R1 = n0 + m0;    R2 = n1 + m1;

   if(C1==0 || C2==0 || R1==0 || R2==0 ) return 0;

   N = R1 + R2;

   n0e = C1*R1/N;   n1e = C1*R2/N;

   m0e = C2*R1/N;   m1e = C2*R2/N;

   chi = (n0-n0e)*(n0-n0e)/n0e + (n1-n1e)*(n1-n1e)/n1e +
         (m0-m0e)*(m0-m0e)/m0e + (m1-m1e)*(m1-m1e)/m1e;

   return chi;
}
//________________________________
// логарифм правдоподобия
double claster::lnLambda(int r1, int r2){
   double sum=0;
   if(r1>0){
      sum += r1*log(r1);
   }
   if(r2>0){
      sum += r2*log(r2);
   }
   sum -= (r1+r2)*log(r1+r2);
   return sum;
}
//_____________________________________________
// первоначальное заполнение списка кластеров cls
// вычисление хи-квадрат расстояния между соседними кластерами
// сортировка pUnitN по убыванию отношения частот случай/контроль
void claster::sortClaster(){
// отсортировать массив 1
   if(fnum1) sort(pUnit1, pUnit1+fnum1, order1());
// отсортировать массив 2
   if(fnum2) sort(pUnit2, pUnit2+fnum2, order2());
// отсортировать массив 3
   if(fnum3) sort(pUnit3, pUnit3+fnum3, order3());
// слить отсортированные строки в pUnit1
   int j;
   for(int i=0; i<fnum2; i++){
      j = fnum1 + i;
      pUnit1[j].ind = pUnit2[i].ind;
      pUnit1[j].r1 = pUnit2[i].r1;
      pUnit1[j].r2 = pUnit2[i].r2;
   }
   for(int i=0; i<fnum3; i++){
      j = fnum1+fnum2+i;
      pUnit1[j].ind = pUnit3[i].ind;
      pUnit1[j].r1 = pUnit3[i].r1;
      pUnit1[j].r2 = pUnit3[i].r2;
   }
   // копирование всех строк в pUnit2
   for(int j=0; j<fnum1+fnum2+fnum3; j++){
      pUnit2[j].ind = j;
      pUnit2[j].r1 = pUnit1[j].r1;
      pUnit2[j].r2 = pUnit1[j].r2;
      pUnit2[j].chi2 = 0;
      pUnit2[j].lamb = lnLambda(pUnit2[j].r1, pUnit2[j].r2);
      // начальное формирование кластеров
      cls.push_back(pUnit2[j]);
   }
   list<fUnit>::iterator it, itlast = cls.end();
   --itlast;// кроме последнего
   int i=0;
   // вычисление хи-квадрат между соседними кластерами
   for(it=cls.begin(); it != itlast; it++){
      double tmp = chi2dist(it);
      qDebug()<<" i="<<i<<" r1="<<(*it).r1<<" r2="<<(*it).r2
      <<" chi2dist="<<tmp;
      (*it).chi2 = tmp;
      i++;
   }
// отладка ===================
//   for(int i=0; i<fnum1+fnum2+fnum3; i++){
//      qDebug()<<" "<<pUnit1[i].ind;
//   }
//   qDebug()<<'\n';
//   for(int i=0; i<fnum1+fnum2+fnum3; i++){
//      qDebug()<<" "<<pUnit1[i].r1;
//   }
//   qDebug()<<'\n';
//   for(int i=0; i<fnum1+fnum2+fnum3; i++){
//      qDebug()<<" "<<pUnit1[i].r2;
//   }
}
//________________________________
// вычисление значимости набора кластеров
// df - кол-во кластеров
double claster::znach(int df){
   double chi2sum = log0;
   list<fUnit>::iterator it;
   for(it=cls.begin(); it != cls.end(); it++){
      chi2sum += (*it).lamb;
//      qDebug()<<" i="<<i<<" r1="<<(*it).r1<<" r2="<<(*it).r2
//      <<" chi2dist="<<tmp;
//      (*it).chi2 = tmp;
//      i++;
   }
   double pVal = NR::chi2distr(2*chi2sum, df);
   return pVal;
}
//==================================================
// слияние ближайших кластеров. Сохранение номера кластера в pUnit2
// возврат номера 1-ого кластера для объединения
int claster::merge(int numclus){
    // пропустить первый шаг с числом кластеров = числу факторов
   if(numclus==NumFactors) return -1;
   qDebug()<<"number of clusters="<<numclus;
   list<fUnit>::iterator it=cls.begin(), it2 = cls.end();
   --it2;// без последнего элемента
   int i=1,// текущий индекс
    imin=0;// индекс первого (из 2-х) объединяемого кластера
   double min = (*it).chi2;// минимальное расстояние хи-квадрат
   for(++it; it != it2; ++it){
      double dist = (*it).chi2;
      if(dist<min){
         min = dist;
         imin = i;
      }
      qDebug()<<" i="<<i<<" dist="<<dist<<" min="<<
      min<<" imin="<<imin;
      i++;
   }
   qDebug()<<" imin="<<imin;
    // объединить 2 кластера =============
   it = cls.begin();// первый кластер
   for(int i=0; i<imin; ++i){
      ++it;
   }
   it2 = it;
   ++it2;       // второй кластер
   // объединить два кластера
   (*it).r1 += (*it2).r1;
   (*it).r2 += (*it2).r2;
   cls.erase(it2);// удалить второй кластер
   (*it).lamb = lnLambda((*it).r1, (*it).r2);
    // пересчитать расстояние после
   it2 = it;
   if(++it2 != cls.end()){// не последний
      (*it).chi2 = chi2dist(it);
      qDebug()<<" right dist="<<(*it).chi2;
   }
   // пересчитать расстояние до
   if(it != cls.begin()){
      --it;
      (*it).chi2 = chi2dist(it);
      qDebug()<<" left dist="<<(*it).chi2;
   }

// отладка ================
   it=cls.begin();
   it2 = cls.end();
   i=0;
   for(it; it != it2; ++it){
      int r1 = (*it).r1;
      int r2 = (*it).r2;

      qDebug()<<" i="<<i<<" r1="<<r1<<" r2="<<r2;
      i++;
   }
    // поместить номер кластера в pUnit2
   int i0=0, j2;
   for( it = cls.begin(); ; ++i0){
      int j1 = (*it).ind;
      ++it;
      if(it == cls.end()) break;// выход из цикла
      j2 = (*it).ind;
      int d = j2-j1;
      for(i=j1; i-j1<d; i++){
         pUnit2[i].ind = i0;
      }
   }
   for(i=j2; i<NumFactors; i++){
      pUnit2[i].ind = i0;
   }
   return imin;
}
//__________________________________________________________
// Сохранение принадлежности факторов кластерам в контейнере
//  numcl = кол-во кластеров, начиная с NumFactors, до 1
//  mincl = кол-во кластеров с Pval=min
void claster::outres(int numcl, int mincl){
    // номер строки имен факторов
//    int i0 = 3 + scell->toprow - 1 + numcl;
    int k;// номер кластера (от 1)
//    QString txt;
    for(int j=0; j<NumFactors; j++){
        k = pUnit2[j].ind + 1;
        clsInd[NumFactors*(numcl-1)+j] = k;
//        txt = "cluster " + QObject::tr("%1").arg(k);
//        spreadsheet->setFormula(i0, j+col0+scell->leftcol, txt);
    }
//    txt.setNum(pval[numcl]);
//    spreadsheet->setFormula(i0, -1+col0+scell->leftcol, txt);

//    if(numcl==2){// последняя итерация
//        // Копирование названия отсортированных факторов вниз таблицы
//        i0 = 3 + scell->toprow;// номер строки факторов
//        for(int j=0; j<NumFactors; j++){// по столбцам
//            int i = pUnit1[j].ind;
//            txt = spreadsheet->item(-1+scell->toprow, i+scell->leftcol)->text();
//            spreadsheet->setFormula(i0, j+col0+scell->leftcol, txt);
//        }// j
//        spreadsheet->setFormula(i0, -1+col0+scell->leftcol, "P-VALUE");
//        // заменить слово cluster на CLUSTER для строки mincl
//        i0 = 3 + scell->toprow - 1 + mincl;
//        for(int j=0; j<NumFactors; j++){
//            txt = spreadsheet->item(i0, j+col0+scell->leftcol)->text();
//            txt.replace(QString("cluster"), QString("CLUSTER"));
//            spreadsheet->setFormula(i0, j+col0+scell->leftcol, txt);
//        }
//    }
}
//___________________________________
// вывод результатов в окно Result
// ВХОД: mincl - кол-во кластеров с p=p_min
void claster::setRes(int mincl){
    // формирование таблицы =====================
    QString ss, sn;
    int numCol = 1;// 1 столбец
//
    ss += "<TR>";
    ss += "<TD>";
    ss += QObject::tr("P-value");
    ss += "</TD>";
    for(int j=0; j<numCol; ++j){
        ss += "<TD>";
        sn.setNum(pval[mincl]);// значимость набора кластеров
        ss += sn;
        ss += "</TD>";
    }
    ss += "</TR>";
    //
    ss += "<TR>";
    ss += "<TD>";
    ss += QObject::tr("Number of clusters");
    ss += "</TD>";
    for(int j=0; j<numCol; ++j){
        ss += "<TD align=center>";
        sn.setNum(mincl);// кол-во кластеров
        ss += sn;
        ss += "</TD>";
    }
    ss += "</TR>";
    //
    for(int i=0; i<NumFactors; ++i){
        ss += "<TR>";
        ss += "<TD>";
        int ind = pUnit1[i].ind;
        ss += scell->factNames.at(ind);
        ss += "</TD>";

        for(int j=0; j<numCol; ++j){
            ss += "<TD>";
            int r0 = mincl-1;
            sn.setNum(clsInd[NumFactors*r0+i]);
            ss += sn;
            ss += "</TD>";
        }
        ss += "</TR>";
    }

//    tblres = "<center><br>Equal risk clusters<TABLE border=1 cellpadding=5>" + ss;
//    tblres += "</TABLE></center>";
    tblres = "<br><br><b>Equal risk clusters</b><br><TABLE border=1 cellpadding=5>" + ss;
    tblres += "</TABLE>";

}
//________________________________
// вывод отношения шансов
void claster::OddsRatio(int numClusters){
    // обнуление
    for(int i=0; i<NumFactors; ++i){
        pUnit2[i].r1 = 0;// численность кластера (от 0) выборки 1
        pUnit2[i].r2 = 0;// численность кластера (от 0) выборки 2
        pUnit2[i].chi2 = 0;// отношение частот
        pUnit2[i].lamb = 0;// отношение шансов
    }

    int r0 = numClusters-1;
    int curCls = clsInd[NumFactors*r0 + 0];// номер текущего кластера (=1)
    for(int i=0; i<NumFactors; ++i){
        if(clsInd[NumFactors*r0 + i] != curCls){
            curCls = clsInd[NumFactors*r0 + i];// следующий кластер
        }
        pUnit2[curCls-1].r1 += pUnit1[i].r1;
        pUnit2[curCls-1].r2 += pUnit1[i].r2;
    }
    double ratioSUM = ((double) sum1)/sum2;
    for(int i=0; i<numClusters; ++i){
        double rr1, rr2;
        if(pUnit2[i].r1 > 0 ){
            rr1 = pUnit2[i].r1;
        }else{
            rr1 = -0.5;
        }
        // отношение частот (<0, когда знаменатель=0)
        pUnit2[i].chi2 = (ratioSUM * pUnit2[i].r2)/rr1;

        if(i < numClusters-1){// отношение шансов
            if(pUnit2[i+1].r2 > 0 ){
                rr2 = pUnit2[i+1].r2;
            }else{
                if(rr1>0){
                    rr2 = -0.5;
                }else{
                    rr2 = 0.5;
                }
            }
            // отношение шансов (<0, когда знаменатель=0)
            pUnit2[i].lamb = (pUnit2[i].r2*pUnit2[i+1].r1)/(rr1*rr2);
        }
    }
    outOdds(numClusters);// вывод отношения шансов
}
//________________________________________
// вывод таблицы
void claster::outOdds(int numClusters){
// формирование таблицы =====================
    QString ss, sn;
    int numCol = 5;// кол-во столбцов
    // Первая строка
    ss += "<TR>";
    ss += "<TD>";
    ss += QObject::tr("Cluster");
    ss += "</TD>";
    ss += "<TD>";
//    ss += QObject::tr("Control");
    ss += scell->sampNames[0];
    ss += "</TD>";
    ss += "<TD>";
//    ss += QObject::tr("Case");
    ss += scell->sampNames[1];
    ss += "</TD>";
    ss += "<TD>";
    ss += QObject::tr("Frequencies ratio");
    ss += "</TD>";
    ss += "<TD>";
    ss += QObject::tr("Odds ratio");
    ss += "</TD>";
    ss += "</TR>";
    for(int i=0; i<numClusters; ++i){
        ss += "<TR>";

        ss += "<TD>";
        sn.setNum(i+1);// номер кластера
        ss += sn;
        ss += "</TD>";
        ss += "<TD>";
        sn.setNum(pUnit2[i].r1);// контроль
        ss += sn;
        ss += "</TD>";
        ss += "<TD>";
        sn.setNum(pUnit2[i].r2);// случай
        ss += sn;
        ss += "</TD>";
        ss += "<TD>";
        if(pUnit2[i].chi2 >= 0){
            sn.setNum(pUnit2[i].chi2);// отношение частот
            ss += sn;
        }else{
            ss += "(";
            sn.setNum(-pUnit2[i].chi2);
            ss += sn;
            ss += ")";
        }
        ss += "</TD>";
        ss += "<TD>";
        if(i<numClusters-1){
            if(pUnit2[i].lamb >= 0){
                sn.setNum(pUnit2[i].lamb);// отношение шансов
                ss += sn;
            }else{
                ss += "(";
                sn.setNum(-pUnit2[i].lamb);
                ss += sn;
                ss += ")";
            }
        }
        ss += "</TD>";
        ss += "</TR>";
    }

//    tblres += "<center><br>Odds and Frequencies Ratio<TABLE border=1 cellpadding=5>" + ss;
//    tblres += "</TABLE></center>";
    tblres += "<br><br><b>Odds and Frequencies Ratio</b><br><TABLE border=1 cellpadding=5>" + ss;
    tblres += "</TABLE>";

}
//________________________________
//    НЕ ИСПОЛЬЗУЕТСЯ !!!
// вывод результатов в электронную таблицу
//  numcl = кол-во кластеров,
//  mincl = кол-во кластеров min(Pval)
void claster::outresSheet(int numcl, int mincl){
    // номер строки имен факторов
    int i0 = 3 + scell->toprow - 1 + numcl;
      int k;// номер кластера
      QString txt;
      for(int j=0; j<NumFactors; j++){
         k = pUnit2[j].ind + 1;
         txt = "cluster " + QObject::tr("%1").arg(k);
         spreadsheet->setFormula(i0, j+col0+scell->leftcol, txt);
      }
      txt.setNum(pval[numcl]);
      spreadsheet->setFormula(i0, -1+col0+scell->leftcol, txt);

    // заменить слово cluster на CLUSTER для mincl
   if(numcl==2){// последняя итерация
       // Копирование названия отсортированных факторов вниз таблицы
       i0 = 3 + scell->toprow;// номер строки факторов
       for(int j=0; j<NumFactors; j++){// по столбцам
          int i = pUnit1[j].ind;
           txt = spreadsheet->item(-1+scell->toprow, i+scell->leftcol)->text();
          spreadsheet->setFormula(i0, j+col0+scell->leftcol, txt);
       }// j
       spreadsheet->setFormula(i0, -1+col0+scell->leftcol, "P-VALUE");
      i0 = 3 + scell->toprow - 1 + mincl;
      for(int j=0; j<NumFactors; j++){
         txt = spreadsheet->item(i0, j+col0+scell->leftcol)->text();
         txt.replace(QString("cluster"), QString("CLUSTER"));
         spreadsheet->setFormula(i0, j+col0+scell->leftcol, txt);
      }
   }
}
