#include <QDebug>
#include <math.h>
#include "fexact2.h"
//____________________________
// конструктор
//   nrow - Maximum number of rows (Input)
//   ncol - Maximum number of columns (Input)
//   fact - logarithms of factorials (Input)
fexact::fexact(int nrow, int ncol, double *fact){
   lnFac = fact;
   // tol is chosen as the square root of the smallest relative spacing
   tol = 3.45254e-7;// допустимое отклонение
   int nco = MAX(nrow, ncol);
   int i = nrow + ncol + 1;
   i40 = 0;
//   i41 = 1;
   i41 = i40 + ncol;
   i42 = i41 + i;
   i43 = i42 + i;
   i44 = i43 + i;
   i45 = i44 + i;
   i46 = i45 + i;
   i47 = i46 + i*nco;
   // + (i * nco)
   // всего = 1+(nrow+ncol+1)[5+2*nco]
   iwk = new int [1+(nrow+ncol+1)*(5+2*nco)+ncol];
   col = iwk;
   Ystk  = new double [nrow+ncol+1];
}
//=========================================
// вычисление кратчайшего пути для таблицы сопряженности
// ВХОД
//    nrow=2 - кол-во срок
//    *irow - сумма строк
//    ncol - кол-во столбцов
//    *icol - сумма столбцов (отсортирован по убыванию)
// ВЫХОД (возврат)
//    return = SUM[ln(Cj!)] - SUM[ln(Yij!)]
//=========================================
double fexact::getSP(int nrow, int *irow, int ncol, int *icol){
   double dspt = 0, lnClmn;
//   int tot=0;
   for(int i=0; i<ncol; i++){
      dspt += lnFac[icol[i]];
      col[ncol-i-1] = icol[i];// сортировка по возрастанию
   }
   lnClmn = dspt;

   for(int i=0; i<nrow; i++){// nrow=2
      dspt += lnFac[irow[i]];
//      tot += irow[i];
   }
   if(irow[0] <= irow[1]){// сортировка по возрастанию
      row[0] = irow[0];
      row[1] = irow[1];
   }else{
      row[1] = irow[0];
      row[0] = irow[1];
   }

//   qDebug()<<" dspt="<<dspt;
//   qDebug()<<" dspt="<<dspt;
//   dspt -= lnFac[tot];
//   double y, pval;
//   double y = lnFac[13]+lnFac[2]+lnFac[9]+lnFac[5];
//   qDebug()<<" y="<<y;
//   pval = dspt - y;
//   qDebug()<<" ln(SP)="<<pval<<" y="<<y;
//   dspt += 100;  // сдвиг p~1.e-100
//   dspt = 2*lnFac[tot];  // сдвиг p~1.e-100
//   qDebug()<<" fact[tot]="<<lnFac[tot];

   double SP = f4xact( nrow, row, ncol, col, dspt,
      lnFac, &iwk[i47], &iwk[i41], &iwk[i42], &iwk[i43],
      &iwk[i44], &iwk[i45], &iwk[i46], Ystk, &tol);
//   double pval = exp(dspt+SP-lnFac[tot]);
//   double y = lnFac[6]+lnFac[4]+lnFac[3];
//   qDebug()<<" Pval="<<pval<<" dspt="<<dspt<<" tot="
//   <<tot<<" nrow="<<nrow<< " ncol="<<ncol
//   <<" lnFac[tot]="<<lnFac[tot]<<" y="<<-y
//   ;
//   return SP;
   if(SP > tol){// error
      qDebug()<<" getSP error calculation SP";
      return -1;
   }
   double res = lnClmn + SP;
//   if(res>0 && res<tol) res=0;
//   qDebug()<<" dspt"<<dspt<<" lnClmn="<<lnClmn<<" lnSP="<<SP
//          <<" res="<<res;
   return res;
}
//____________________________
// деструктор
fexact::~fexact(){
   delete [] iwk;
   delete [] Ystk;
}
//==================================================
// ИСПРАВЛЕНО: if (nrow * ncol == 4) {}
// ----------------------------------------
//  Name:	      F4XACT
//  Purpose:    Computes the shortest path length for a given table.

//  Arguments:
//     NROW   - The number of rows in the table.	(Input)
//     IROW   - Vector of length NROW containing the row sums for the
//	      table.  (Input)
//     NCOL   - The number of columns in the table.  (Input)
//     ICOL   - Vector of length K containing the column sums for the
//	      table.  (Input)
//     DSPT   - "offset"  for SP computation
//     FACT   - Vector containing the logarithms of factorials.  (Input)
//     ICSTK  - NCOL by MAX(NROW,NCOL) work array.
//     NCSTK  - Work vector of length NROW+NCOL+1.
//     LSTK   - Work vector of length NROW+NCOL+1.
//     MSTK   - Work vector of length NROW+NCOL+1.
//     NSTK   - Work vector of length NROW+NCOL+1.
//     NRSTK  - Work vector of length NROW+NCOL+1.
//     IRSTK  - NROW by MAX(NROW,NCOL) work array.
//     YSTK   - Work vector of length NROW+NCOL+1.
//     TOL    - Tolerance.					(Input)

//  Return Value :

//    SP	    - The shortest path for the table.			(Output)
//  -----------------------------------------------------------------------
double fexact::f4xact(int nrow, int *irow, int ncol, int *icol, double dspt,
   double *fact, int *icstk, int *ncstk, int *lstk, int *mstk,
   int *nstk, int *nrstk, int *irstk, double *ystk,
   const double *tol)
{

   /* Local variables */
   int i, j, k, l, m, n, ic1, ir1, ict, irt, istk, nco, nro;
   double y, amx, SP;

   /* Take care of the easy cases first */
   if (nrow == 1) {
      SP = 0.;
      for (i = 0; i < ncol; ++i)
         SP -= fact[icol[i]];
      return SP;
   }
   if (ncol == 1) {
      SP = 0.;
      for (i = 0; i < nrow; ++i)
         SP -= fact[irow[i]];
      return SP;
   }
   if (nrow * ncol == 4) {
      if (irow[1] <= icol[1]){
         SP = -(fact[irow[1]] + fact[icol[0]] + fact[icol[1] - irow[1]]);
      }else{
         SP = -(fact[icol[1]] + fact[irow[0]] + fact[irow[1] - icol[1]]);
      }
      return SP;
   }

   /* Parameter adjustments */
   irstk -= nrow + 1;
   icstk -= ncol + 1;

   --nrstk;
   --ncstk;
   --lstk;
   --mstk;
   --nstk;
   --ystk;

   /* initialization before loop */
   for (i = 1; i <= nrow; ++i){
//      int tmp = irow[nrow - i];
      irstk[i + nrow] = irow[nrow - i];
   }
   for (j = 1; j <= ncol; ++j){
//      int tmp = icol[ncol - j];
      icstk[j + ncol] = icol[ncol - j];
   }
   nro = nrow;
   nco = ncol;
   nrstk[1] = nro;
   ncstk[1] = nco;
   ystk[1] = 0.;
   y = 0.;
   istk = 1;
   l = 1;
   amx = 0.;
   SP = dspt;

   /* First LOOP */
   do {
      ir1 = irstk[istk * nrow + 1];
      ic1 = icstk[istk * ncol + 1];
      if (ir1 > ic1) {
         if (nro >= nco) {
            m = nco - 1;	n = 2;
         } else {
            m = nro;	n = 1;
         }
      } else if (ir1 < ic1) {
         if (nro <= nco) {
            m = nro - 1;	n = 1;
         } else {
            m = nco;	n = 2;
         }
      } else {
         if (nro <= nco) {
            m = nro - 1;	n = 1;
         } else {
            m = nco - 1;	n = 2;
         }
      }

L60:
		if (n == 1) {
			i = l; j = 1;
		} else {
			i = 1; j = l;
		}

		irt = irstk[i + istk * nrow];
		ict = icstk[j + istk * ncol];
//		y += fact[imin2(irt, ict)];
		y += fact[MIN(irt, ict)];
		if (irt == ict) {
			--nro;
			--nco;
			f11act(&irstk[istk * nrow + 1], i, nro,
					 &irstk[(istk + 1) * nrow + 1]);
			f11act(&icstk[istk * ncol + 1], j, nco,
					 &icstk[(istk + 1) * ncol + 1]);
		} else if (irt > ict) {
			--nco;
			f11act(&icstk[istk * ncol + 1], j, nco,
					 &icstk[(istk + 1) * ncol + 1]);
			f8xact(&irstk[istk * nrow + 1], irt - ict, i, nro,
					 &irstk[(istk + 1) * nrow + 1]);
		} else {
			--nro;
			f11act(&irstk[istk * nrow + 1], i, nro,
					 &irstk[(istk + 1) * nrow + 1]);
			f8xact(&icstk[istk * ncol + 1], ict - irt, j, nco,
					 &icstk[(istk + 1) * ncol + 1]);
		}

		if (nro == 1) {
			for (k = 1; k <= nco; ++k){
//				double tmp1 = icstk[k + (istk + 1) * ncol];
//				qDebug()<<" tmp1="<<tmp1;
				y += fact[icstk[k + (istk + 1) * ncol]];
			}
			break;// выход из цикла do-while
		}
		if (nco == 1) {
			for (k = 1; k <= nro; ++k){
//				double tmp2 = irstk[k + (istk + 1) * nrow];
//				qDebug()<<" tmp2="<<tmp2;
				y += fact[irstk[k + (istk + 1) * nrow]];
			}
			break;// выход из цикла do-while
		}

		lstk[istk] = l;
		mstk[istk] = m;
		nstk[istk] = n;
		++istk;
		nrstk[istk] = nro;
		ncstk[istk] = nco;
		ystk[istk] = y;
		l = 1;
	} while(1);/* end do */

	/* L90:*/
	if (y > amx) {
		amx = y;
		if (SP - amx <= *tol)
//			return -dspt;
			return 1;// ошибка
	}

	/* L100: */
	do {
		--istk;
		if (istk == 0) {
			SP -= amx;
			if (SP - amx <= *tol)
//				return -dspt;
				return 1;// ошибка
			else
				return SP - dspt;
		}
		l = lstk[istk] + 1;

		/* L110: */
		for(;; ++l) {
			if (l > mstk[istk])	break;

			n = nstk[istk];
			nro = nrstk[istk];
			nco = ncstk[istk];
			y = ystk[istk];
			if (n == 1) {
				if (irstk[l	+ istk * nrow] <
					 irstk[l - 1 + istk * nrow])	goto L60;
			}
			else if (n == 2) {
				if (icstk[l	+ istk * ncol] <
					 icstk[l - 1 + istk * ncol])	goto L60;
			}
		}
	} while(1);
}
//_______________________________________________________
//  Name:	      F11ACT
//  Purpose:    Routine for revising row totals.

//  Arguments:
//     IROW   - Vector containing the row totals.	(Input)
//     I1	    - Indicator.			(Input)
//     I2	    - Indicator.  			(Input)
//     NEW    - Vector containing the row totals.	(Output)
//_______________________________________________________
void fexact::f11act(int *irow, int i1, int i2, int *new1){
   int i;

   for (i = 0;  i < (i1 - 1); ++i)	new1[i] = irow[i];
   for (i = i1; i <= i2; ++i)      new1[i-1] = irow[i];

   return;
}
//_______________________________________________________
//  Name:	      F8XACT
//  Purpose:    Routine for reducing a vector when there is a zero
//	      element.
//  Arguments:
//     IROW   - Vector containing the row counts.			(Input)
//     IS	    - Indicator.					(Input)
//     I1	    - Indicator.					(Input)
//     IZERO  - Position of the zero.				(Input)
//     NEW    - Vector of new row counts.				(Output)
//  -----------------------------------------------------------------------
void fexact::f8xact(int *irow, int is, int i1, int izero, int *new1){
   int i;

   /* Parameter adjustments */
   --new1;
   --irow;

   /* Function Body */
   for (i = 1; i < i1; ++i)
      new1[i] = irow[i];

   for (i = i1; i <= izero - 1; ++i) {
      if (is >= irow[i + 1])
         break;
      new1[i] = irow[i + 1];
   }

   new1[i] = is;

   for(;;) {
      ++i;
      if (i > izero)
         return;
      new1[i] = irow[i];
   }
}
