#include <math.h> // log()
#include <QDebug>
#include "interact.h"

//___________________________________________
// все вычисления в конструкторе
// ВХОД:
//   colnum (=2) - кол-во столбцов матриц
//   x1 - выб.1 квадратная матрица по строкам
//   x2 - ... выб.2
// ВЫХОД:
//   err - код ошибки (успех =0)
//   pVal - вероятность нулевой гипотезы
interact::interact(int &err, int colnum, int numrows,
   int *x1, int *x2, double &pVal)
{
    cols = colnum;
    rows = numrows;
    xMatr = x1;
    yMatr = x2;

    ln_n0 = 0;
    for(int i=0; i<2; i++){
       Ri[i] = 0;
       Cj[i] = 0;
    }
    Tij = 0;
    lnFact = 0;
    err = init();   // инициализация
    if(err) return; // ошибка
    double pTotal=1, pSuccess=1;
    int xmin = nDevN0(-1, pTotal, pSuccess);
    int xmax = nDevN0(+1, pTotal, pSuccess);
    pVal = pSuccess/pTotal;
}
//_________________________
// расчет вероятности
// Возврат: граничное X_11
// Вход: dir - направление (+/-)1
// ВЫход:
//  Pt - полное отношение
//  P0 - удачное отношение
int interact::nDevN0(int dir, double& Pt,  double& P0){
    int xBound = xMatr[0];
    for(int i=dir; 1; i+=dir){
       double ln;
       int er = ln_n(ln, i);
       if(er) break;

       xBound += i;
       double dlnN0 = ln - ln_n0;
       double dP = exp(dlnN0);
       Pt += dP;
       if(dlnN0 < EPSINT){
          P0 += dP;
       }
    }
    return xBound; //
}
//____________________________________________
// Вычисление количества способов формирования таблиц
// Выход:   ln - логарифм числа способов
// Вход:    dx - приращение x_11
// Возврат:
//    0 - успех
//    1 - таблица не существует
int interact::ln_n(double &ln, int dx){
    ln = 0;
    for(int i=0; i<rows; i++){    // по строкам
        for(int j=0; j<cols; j++){    // по строкам
            int deltaX = dx;
            int k = i*cols+j;
            if((i+j)%2){
                deltaX *= -1;
            }
            bool ok = isExist(deltaX, i, j, k);
            if(!ok) return 1;

            int x = xMatr[k] + deltaX;
            double Cmn = ln_Binom(Tij[k], x);
            ln += Cmn;
        }
    }
    return 0;
}
//________________________________
// действительная конфигурация 2-х таблиц ?
// Вход:
//  dx - приращение элемента (i,j) 1-ой выборки
bool interact::isExist(int dx, int i, int j, int k){
    int x = xMatr[k] + dx;
    // 1-ая выборка
    if(x<0 || x>Tij[k] || x>Cj[0][j] || x>Ri[0][i] )
    {
       return false;
    }
    // 2-ая выборка
    int y = yMatr[k] - dx;
    if(y>Cj[1][j] || y>Ri[1][i])
    {
       return false;
    }
    return true;
}
//_________________________
// деструктор
// освобождаем динамич.память
interact::~interact(){
    for(int i=0; i<2; i++){
        if(Ri[i])  delete [] Ri[i];
        if(Cj[i])  delete [] Cj[i];
    }
    if(lnFact)  delete [] lnFact;

}
//___________________________________________________
// логарифм биномиального коэффициента (m, n), m>=n
double interact::ln_Binom(int m, int n){
   double lnC = lnFact[m] - lnFact[n] - lnFact[m-n];
//   qDebug()<<" m="<<m<<" n="<<n<<" binCoeff="<<expln;
   return lnC;
}
//_________________________
// инициализация
// Возврат: 0 - успех
int interact::init(){
    for(int i=0; i<2; i++){
        Ri[i] = new int[rows];
        Cj[i] = new int[cols];
    }

    Tij = new int[rows*cols];
    if(Tij==0) return 1;// ошибка
    int tMax = maxT();
    lnFact = new double[tMax+1];// lnFact[]=ln(x!)
    if(lnFact==0) return 2;// ошибка
    double logFac = 0;
    lnFact[0] = 0;
    for(int i=1; i<=tMax; i++){
       double loge = log(i);
       logFac += loge;
       lnFact[i] = logFac;
//       qDebug()<<" i="<<i<<" lnFact[i]="<<lnFact[i];
    }
    int err = ln_n(ln_n0, 0);
    return err;
}
//_________________________________________________
// инициализация массивов
// максимум гаплотипов T_ij для вычисления факториалов
int interact::maxT(){
    int max=0; // максимум гаплотипов
    for(int i=0; i<rows; i++){    // по строкам
        int R1=0, R2=0; // сумма строки
        for(int j=0; j<cols; j++){// по столбцам
            int k = i*cols+j;
            R1 += xMatr[k];
            R2 += yMatr[k];
            int t = xMatr[k] + yMatr[k];
            Tij[k] = t;
            if(t>max){
              max = t;
            }
        }
        Ri[0][i] = R1;
        Ri[1][i] = R2;
    }

    for(int j=0; j<cols; j++){// по столбцам
        int C1=0, C2=0; // сумма столбца
        for(int i=0; i<rows; i++){    // по строкам
            int k = i*cols+j;
            C1 += xMatr[k];
            C2 += yMatr[k];
        }
        Cj[0][j] = C1;
        Cj[1][j] = C2;
    }
    return max;
}
