#include <QtGui>
#include <QActionEvent>
#include <ctime>

#include "finddialog.h"
#include "gotocelldialog.h"
#include "mainwindow.h"
#include "sortdialog.h"
#include "spreadsheet.h"
#include "seldialog.h"
#include "nr.h"      // хи-квадрат
#include "nfisher.h" // точный тест Фишера
#include "cluster.h" // кластеризация
#include "wein.h"    // харди-вайнберг
#include "interact.h" // взаимодействие

extern QApplication *gpApplication;// обработка событий
int gStopCalculation;   // внешний флаг Останова
int gProgressCalc;      // прогресс вычислений
//--------------------
int StopCalculation(){
// обработка событий
   if(gpApplication) gpApplication->processEvents();
   return gStopCalculation;
}

MainWindow::MainWindow()
{
    spreadsheet = new Spreadsheet;
    spreadsheet->setStyleSheet(QString(
   "QTableWidget {selection-background-color: lightblue;"
   " selection-color: black}"
    ));

    setCentralWidget(spreadsheet);

    createDockWindows();
    createActions();

    createMenus();
    createContextMenu();
    createToolBars();
    createStatusBar();

    readSettings();
    findDialog = 0;

    setWindowIcon(QIcon(":/images/icon.png"));
    setCurrentFile("");

    setWindowTitle(tr("DoctorStat"));

    setUnifiedTitleAndToolBarOnMac(true);

// создать диалоговое окно выбора ячеек
   pSelDlg = new SelDialog(spreadsheet, this);
   connect(pSelDlg, SIGNAL(accepted()), SLOT(slotCalc()));
   timerCalc = new QTimer(this);
   connect(timerCalc, SIGNAL(timeout()),
      SLOT(slotProgress()));
//   count = 0;
}
//_________________________________
// выделить ячейки
void MainWindow::SelectCells(){
   emit signalTestRange();// тестирование диапазона ячеек
//   pSelDlg->show();
   pSelDlg->showSpec();
}
//____________________________________________
// Харди-Вайнберг
void MainWindow::slotMenuGeneInter(){
   pSelDlg->numTest = GINTER;
   SelectCells();
}
//____________________________________________
// Харди-Вайнберг
void MainWindow::slotMenuHardyW(){
   pSelDlg->numTest = HARDYW;
//   pSelDlg->chk2x2->setHidden(true);
   SelectCells();
}
//____________________________________________
// кластеризация
void MainWindow::slotMenuCluster(){
   pSelDlg->numTest = CLUSTER;
//   pSelDlg->chk2x2->setHidden(true);
   SelectCells();
}
//____________________________________________
// вычисление хи-квадрат
void MainWindow::slotMenuChi2(){
   // дописать!!!
   pSelDlg->numTest = Chi2;
   SelectCells();
}
//____________________________________________
// вычисление критерия Фишера
void MainWindow::slotMenuFisher(){
   pSelDlg->numTest = Fisher;
   SelectCells();
}
//______________________________
// нажата кнопка ОК в окне выделения ячеек таблицы
// вызов функций вычисления
void MainWindow::slotCalc(){
   switch(pSelDlg->numTest){
   case GINTER:   // Взаимодействие
      gInteraction();
      break;
   case Fisher:   // точный критерий фишера
      fisher();
      break;
   case Chi2:     // критерий хи-квадрат
      chi2();
      break;
   case CLUSTER:  // кластеризация
      clusterCalc();
      break;
   case HARDYW:   // Харди-Вайнберг
      hardyWcalc();
      break;
   }
}
//____________________________________________
// вычисление взаимодействия генов
// Дописать !/////
void MainWindow::gInteraction(){

    pSelDlg->numTest = -1;
    // получить данные
    int numClm = pSelDlg->cols;// кол-во столбцов
    int numRows = pSelDlg->rows/2;// кол-во строк
    int *r1 = pSelDlg->R1; // 1-ая выборка
    int *r2 = pSelDlg->R2; // 2-ая выборка
    // вывести исходные данные
    QString xv1;
    for(int k=0; k<numRows*numClm; k++){
        if(k==0 ) xv1 += "<br /> { ";
        xv1 += QString::number(r1[k]) ;
        if(k<numRows*numClm-1){
            xv1 += ", ";
        }else{
            xv1 += " }";
        }
    }
    for(int k=0; k<numRows*numClm; k++){
        if(k==0 ) xv1 += " { ";
        xv1 += QString::number(r2[k]) ;
        if(k<numRows*numClm-1){
            xv1 += ", ";
        }else{
            xv1 += " }";
        }
    }
    int err=0;  // флаг результата
    double pVal;  // p-value
    interact* pInt = new interact(err, numClm, numRows, r1, r2, pVal);
   QString str;
    if(err){
       str = "Interact error=" + QString::number(err);
       qDebug() << str;
     }else{
       qDebug() << "Interact Pvalue="<<pVal;
     }
    delete pInt;

    switch(err){
    case 0:
       str = tr(
          "<br /> Interact %1x%2 test <b>P-VALUE = %3</b>  for input matrix:"
       ).arg(numRows).arg(numClm).arg(pVal) + xv1;
       break;
    default:// ошибка
       str = tr(
       "<br />Program is stopped by user");
       break;///
    }
    outResult(str, err);
}
//____________________________________________
// вычисление равновесия Харди-Вайнберга
void MainWindow::hardyWcalc(){
   pSelDlg->numTest = -1;
// получить векторы
   int numClm = pSelDlg->numCol;
   int *x_vec = pSelDlg->R1;

   QString xv1;
   int k=0;
   for(int i=0; i<numClm; i++){
      xv1 += "<br /> { ";
      for(int j=0; j<numClm; j++){
         k = i*numClm+j;
//      qDebug()<<" i="<<i<<" x="<<x_vec[i]<<" col="<<col_sum[i];
         xv1 += QString::number(x_vec[k]) ;
         if(j<numClm-1){
            xv1 += ", ";
         }else{
            xv1 += " }";
         }
      }
   }
   startCalc(numClm);// индикатор прогресса
   double pVal = 0;  // p-value
   int err=0;  // флаг результата
   time_t tim1 = time(0);
   hwein *pHW = new hwein(err, numClm, x_vec, pVal);
   QString str;
   if(err){
      str = "constructor hardy-weinbrg error=" + QString::number(err);
      qDebug() << str;
//      exit(1);
   }
   time_t dt, tim2 = time(0);
   dt = tim2 - tim1;
   qDebug() << "Pvalue="<<pVal<<" dtime="<<dt<<" sec";
   delete pHW;

// скрыть кнопку останова и индикатор прогресса
   startCalc(-1);

   switch(err){
   case 0:
      str = tr(
         "<br /> Hardy-Weinberg %1x%1 test <b>P-VALUE = %2</b>  for input vectors:"
      ).arg(numClm).arg(pVal) + xv1;
      break;
   default:
      str = tr(
      "<br />Program is stopped by user");
      break;///
   }
   outResult(str, err);
}
//____________________________________________
// вычисления кластеризации по степени риска
void MainWindow::clusterCalc(){
   int isAutoRec = autoRecalcOff();
   pSelDlg->numTest = -1;// не входить в seldialog
   int ProgBar = pSelDlg->scell.cols-1;
   startCalc(ProgBar);// запустить полосу прогресса
   int err=0;
   QString str;// строка результатов в формате HTML
   claster pCls(spreadsheet, &pSelDlg->scell, err, str);
   startCalc(-1);    // остановить полосу прогресса
   if(isAutoRec) autoRecalcOff(1);
   outResult(str, err);
}
//____________________________________________
// вычисление критерия Фишера
void MainWindow::fisher(){
   pSelDlg->numTest = -1;
// получить векторы
   int *x_vec = pSelDlg->xVector;
   int *col_sum = pSelDlg->colSum;
   int numClm = pSelDlg->numCol;
   int tblSum = pSelDlg->total;
   int iNum = pSelDlg->iNum;// точность - кол-во знаков

   QString xv1 = QString("<br />x1 = { "),
   xv2 = QString("<br />x2 = { ");
   QString numDig;
   for(int i=0; i<numClm; i++){
      qDebug()<<" i="<<i<<" x="<<x_vec[i]<<" col="<<col_sum[i];
      xv1 += QString::number(x_vec[i]) ;
      xv2 += QString::number(col_sum[i]-x_vec[i]) ;
      if(i<numClm-1){
         xv1 += ", ";
         xv2 += ", ";
      }
   }
   xv1 += " }";
   xv2 += " }";

// запустить вычисления ------------------
   startCalc(numClm);

   int err=0;// флаг результата
   QString str="All is OK!";

  dfex<double> *pFE;
  dfish<path> *pFA;
   QString typTest;
  if(iNum==0){  // точный
     numDig = "";
     typTest = "Exact";
     pFE = new dfex<double>(err, tblSum);
  }else{  // приближенный
     numDig = tr("<br/>Accuracy = %1 digits after point").arg(iNum);
     typTest = "Approximate";
     pFA = new dfish<path>(err, tblSum, iNum);
  }
  if(err){
     str = "constructor nfisher error=" + QString::number(err);
     qDebug() << str;
//     exit(1);
  }

   time_t tim1 = time(0);
   double pVal = 0;// p-value
   if(!err){
      if(iNum==0){
         err = pFE->getPvalue(x_vec, col_sum, numClm, pVal);
      }else{
         err = pFA->getPvalue(x_vec, col_sum, numClm, pVal);
      }
   }
   time_t dt, tim2 = time(0);
   dt = tim2 - tim1;
   qDebug() << "Pvalue="<<pVal<<" dtime="<<dt<<" sec";

   if(iNum==0){
     delete pFE;
   }else{
     delete pFA;
   }

// скрыть кнопку останова и индикатор прогресса
   startCalc(-1);

   switch(err){
   case 0:
      str = tr(
         "<br />%3 Fisher 2x%2 test <b>P-VALUE = %1</b>  for input vectors:"
      ).arg(pVal).arg(numClm).arg(typTest) + numDig + xv1 + xv2;
      break;
   case 10:
      str = tr(
      "<br />Program is stopped by user");
      break;///
   default:
      str = tr(
      "<br />Not enouph memory error=%1").arg(err);
   }
   outResult(str, err);
}
//____________________________________________
// вычисление хи-квадрат
void MainWindow::chi2(){
   double chi2sum = pSelDlg->chi2sum;
   int df = pSelDlg->numCol;
   int rows = pSelDlg->rows;
   int cols = pSelDlg->cols;
   int toprow = pSelDlg->toprow;
   int leftcol = pSelDlg->leftcol;
   qDebug() << "chi2sum="<<chi2sum<<" df="<<df;
   QString xv="";

   for(int i=0; i<rows; i++){
      xv += "<br />{ ";
      for(int j=0; j<cols; j++){
         QTableWidgetItem *item =
            spreadsheet->item(i+toprow, j+leftcol);
         xv += item->text();
         if(j<cols-1){
            xv += ", ";
         }
      }// j
      xv += " }";
   }// i
   xv += tr("<br />chi2=%1 , df=%2").arg(chi2sum).arg(df);
   int err=0;
   double pVal = NR::chi2distr(chi2sum, df);//
   if(pVal<0) err=1;// ошибка
   QString str;
   switch(err){
   case 0:
      str = tr(
            "<br />Chi-squared test <b>P-VALUE = %1</b>  for input vectors:"
            ).arg(pVal) + xv;
      break;
   default:
      str = tr(
            "<br />Programm error=%1").arg(err);
   }
   outResult(str, err);
   pSelDlg->numTest = -1;
   pSelDlg->chk2x2->setHidden(true);
}
//____________________________________________
// Показать индикатор прогресса и запустить таймер
void MainWindow::startCalc(int ProgBar){
   if(ProgBar>0){// Запустить вычисления
      gStopCalculation = 0;
      gProgressCalc = 0;
      pbCalc->reset();
      pbCalc->setRange(0, ProgBar);
      lbIndent->show();
      pbCalc->show();
      btnStop->show();
      timerCalc->start(1000);
   }else{   // остановить
      lbIndent->hide();
      btnStop->hide();
      pbCalc->hide();
      timerCalc->stop();
   }
}
//_________________________________________
// Вывод результата вычислений в окно Result
// Вход:
//  strM - строка результата
//  err - цвет вывода: 0 - синий, != 0 - красный
void MainWindow::outResult(QString &strM, int err){
   QString str;
   str = lbResult->text();// получить предыдущий текст
   str.replace(QString("color=\"darkblue\">"),QString("color=\"darkgreen\">"));
   str.replace(QString("color=\"red\">"),QString("color=\"darkgreen\">"));
   QString strE = tr("</font>");
   QString strB;
   if(err==0 ){
      strB = tr("<font size=\"+1\" color=\"darkblue\">");
   }else{// err!=0
      strB = tr("<font size=\"+1\" color=\"red\">");
   }
   strB += tr(
   "<br /><b>_________________________________________________________</b>"
   );
   str += strB + strM + strE;
   lbResult->setText(str);
   lbResult->adjustSize();
   int ymax = lbResult->height();
   showDock();  // показать окно результатов
   saResult->ensureVisible(0, ymax, 0, 0);// прокрутить окно вниз
}
//_______________________________
// показать прогресс вычислений
void MainWindow::slotProgress(){
   pbCalc->setValue(gProgressCalc);
}
//______________________________________
// создать dock-окно для вывода результатов
void MainWindow::createDockWindows()
{
    dock = new QDockWidget(tr("Results"), this);
    dock->setAllowedAreas(Qt::BottomDockWidgetArea);
    dock->setFeatures(QDockWidget::DockWidgetClosable);

   lbResult = new QLabel(dock);
   lbResult->setAlignment(Qt::AlignTop);
   lbResult->setTextInteractionFlags(Qt::TextBrowserInteraction);
   lbResult->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Ignored);
   lbResult->setScaledContents(true);
//   QString strB = tr("<font size=\"+2\">");
//   QString strE = tr("</font>");
//   QString str1 = strB + tr(
//            "<br /><br />"
//            ) + strE;
//   lbResult->setText(str1);
   lbResult->setStyleSheet(QString(
     "QLabel { background-color: white }"));

   saResult = new QScrollArea();
   saResult->setVerticalScrollBarPolicy(Qt::ScrollBarAsNeeded);
   saResult->setWidget(lbResult);
   saResult->setStyleSheet(QString(
   "QScrollArea { background: white }"));
   dock->setWidget(saResult);
   addDockWidget(Qt::BottomDockWidgetArea, dock);
}
//____________________________________________
// что эта функция делает?
void MainWindow::closeEvent(QCloseEvent *event)
{
    if (okToContinue()) {
        writeSettings();
        event->accept();
    } else {
        event->ignore();
    }
}

void MainWindow::newFile()
{
    if (okToContinue()) {
        spreadsheet->clear();
        setCurrentFile("");
    }
}

void MainWindow::open()
{
    if (okToContinue()) {
        QString fileName = QFileDialog::getOpenFileName(this,
         tr("Open Spreadsheet"), ".",
         tr("Spreadsheet files (*.sp)"));
        if (!fileName.isEmpty())
            loadFile(fileName);
    }
}

bool MainWindow::save()
{
    if (curFile.isEmpty()) {
        return saveAs();
    } else {
        return saveFile(curFile);
    }
}

bool MainWindow::saveAs()
{
    QString fileName = QFileDialog::getSaveFileName(this,
                               tr("Save Spreadsheet"), ".",
                               tr("Spreadsheet files (*.sp)"));
    if (fileName.isEmpty())
        return false;

    return saveFile(fileName);
}

void MainWindow::find()
{
    if (!findDialog) {
        findDialog = new FindDialog(this);
        connect(findDialog, SIGNAL(findNext(const QString &,
                                            Qt::CaseSensitivity)),
                spreadsheet, SLOT(findNext(const QString &,
                                           Qt::CaseSensitivity)));
        connect(findDialog, SIGNAL(findPrevious(const QString &,
                                                Qt::CaseSensitivity)),
                spreadsheet, SLOT(findPrevious(const QString &,
                                               Qt::CaseSensitivity)));
    }

    findDialog->show();
    findDialog->raise();
    findDialog->activateWindow();
}
//_________________________________
// выключить autoRecalc
int MainWindow::autoRecalcOff(int yes){
   int isAutoRecalc = 0;
   if(yes || autoRecalcAction->isChecked()){
      QChar ch = 65 + 'R' - 'A';  // R
      int nKey = Qt::Key_A + 'R' - 'A';
      QKeyEvent *ae = new QKeyEvent(QEvent::KeyPress,
         nKey, Qt::ControlModifier, ch);
      QApplication::sendEvent(this, ae);
      isAutoRecalc = 1;
   }
   return isAutoRecalc;
}
//_________________________________
// показать окно результатов dock
void MainWindow::showDock(){
   if(!showResult->isChecked()){
      QChar ch = 65+1;  // B
      int nKey = Qt::Key_A + 1;
      QKeyEvent *ae = new QKeyEvent(QEvent::KeyPress,
         nKey, Qt::ControlModifier, ch);
      QApplication::sendEvent(this, ae);
   }
}
//_________________________________
void MainWindow::goToCell()
{
//   showDock();
   GoToCellDialog dialog(this);
    if (dialog.exec()) {
        QString str = dialog.lineEdit->text().toUpper();
        spreadsheet->setCurrentCell(str.mid(1).toInt() - 1,
                                    str[0].unicode() - 'A');
    }
}

void MainWindow::sort()
{
    SortDialog dialog(this);
    QTableWidgetSelectionRange range = spreadsheet->selectedRange();
    dialog.setColumnRange('A' + range.leftColumn(),
                          'A' + range.rightColumn());

    if (dialog.exec()) {
        SpreadsheetCompare compare;
        compare.keys[0] =
              dialog.primaryColumnCombo->currentIndex();
        compare.keys[1] =
              dialog.secondaryColumnCombo->currentIndex() - 1;
        compare.keys[2] =
              dialog.tertiaryColumnCombo->currentIndex() - 1;
        compare.ascending[0] =
              (dialog.primaryOrderCombo->currentIndex() == 0);
        compare.ascending[1] =
              (dialog.secondaryOrderCombo->currentIndex() == 0);
        compare.ascending[2] =
              (dialog.tertiaryOrderCombo->currentIndex() == 0);
        spreadsheet->sort(compare);
    }
}

void MainWindow::openRecentFile()
{
    if (okToContinue()) {
        QAction *action = qobject_cast<QAction *>(sender());
        if (action)
            loadFile(action->data().toString());
    }
}

void MainWindow::updateStatusBar()
{
    locationLabel->setText(spreadsheet->currentLocation());
    formulaLabel->setText(spreadsheet->currentFormula());
}

void MainWindow::spreadsheetModified()
{
    setWindowModified(true);
    updateStatusBar();
}
//__________________________________________
// остановить вычисления
void MainWindow::slotStopCalc(){
   gStopCalculation = 1;
}
//__________________________________________
void MainWindow::createActions()
{
    gInter = new QAction(tr("Gene interaction..."), this);
    gInter->setStatusTip(tr("Test Gene interaction"));
//    hardyW->setShortcut(tr("Ctrl+F6"));
    connect(gInter, SIGNAL(triggered()),
       SLOT(slotMenuGeneInter()));

    hardyW = new QAction(tr("Exact Hardy-Weinberg..."), this);
   hardyW->setStatusTip(tr("Test Hardy-Weinberg equilibrium"));
   hardyW->setShortcut(tr("Ctrl+F6"));
   connect(hardyW, SIGNAL(triggered()),
      SLOT(slotMenuHardyW()));

   cluster = new QAction(tr("Risk cluster..."), this);
   cluster->setStatusTip(tr("Risk clusterization"));
   cluster->setShortcut(tr("Ctrl+F7"));
   connect(cluster, SIGNAL(triggered()),
      SLOT(slotMenuCluster()));

   statChi2 = new QAction(tr("&Chi-squared test..."), this);
   statChi2->setStatusTip(tr("Calculate chi-squared test"));
   statChi2->setShortcut(tr("Ctrl+F8"));
   connect(statChi2, SIGNAL(triggered()),
      SLOT(slotMenuChi2()));

   statFisher = new QAction(tr("&Fisher 2xN test..."), this);
   statFisher->setStatusTip(tr("Calculate Fisher Test"));
   statFisher->setShortcut(tr("Ctrl+F9"));
   connect(statFisher, SIGNAL(triggered()),
      SLOT(slotMenuFisher()));

   showResult = dock->toggleViewAction();
    showResult->setCheckable(true);
    showResult->setText(tr("Show &Results"));
    showResult->setStatusTip(tr("Show result window"));
    showResult->setShortcut(tr("Ctrl+B"));

    newAction = new QAction(tr("&New"), this);
    newAction->setIcon(QIcon(":/images/new.png"));
    newAction->setShortcut(QKeySequence::New);
    newAction->setStatusTip(tr("Create a new spreadsheet file"));
    connect(newAction, SIGNAL(triggered()), this, SLOT(newFile()));

    openAction = new QAction(tr("&Open..."), this);
    openAction->setIcon(QIcon(":/images/open.png"));
    openAction->setShortcut(QKeySequence::Open);
    openAction->setStatusTip(tr("Open an existing spreadsheet file"));
    connect(openAction, SIGNAL(triggered()), this, SLOT(open()));

    saveAction = new QAction(tr("&Save"), this);
    saveAction->setIcon(QIcon(":/images/save.png"));
    saveAction->setShortcut(QKeySequence::Save);
    saveAction->setStatusTip(tr("Save the spreadsheet to disk"));
    connect(saveAction, SIGNAL(triggered()), this, SLOT(save()));

    saveAsAction = new QAction(tr("Save &As..."), this);
    saveAsAction->setStatusTip(tr("Save the spreadsheet under a new "
                                  "name"));
    connect(saveAsAction, SIGNAL(triggered()), this, SLOT(saveAs()));

    for (int i = 0; i < MaxRecentFiles; ++i) {
        recentFileActions[i] = new QAction(this);
        recentFileActions[i]->setVisible(false);
        connect(recentFileActions[i], SIGNAL(triggered()),
                this, SLOT(openRecentFile()));
    }

    exitAction = new QAction(tr("E&xit"), this);
    exitAction->setShortcut(tr("Ctrl+Q"));
    exitAction->setStatusTip(tr("Exit the application"));
    connect(exitAction, SIGNAL(triggered()), this, SLOT(close()));

    cutAction = new QAction(tr("Cu&t"), this);
    cutAction->setIcon(QIcon(":/images/cut.png"));
    cutAction->setShortcut(QKeySequence::Cut);
    cutAction->setStatusTip(tr("Cut the current selection's contents "
                               "to the clipboard"));
    connect(cutAction, SIGNAL(triggered()), spreadsheet, SLOT(cut()));

    copyAction = new QAction(tr("&Copy"), this);
    copyAction->setIcon(QIcon(":/images/copy.png"));
    copyAction->setShortcut(QKeySequence::Copy);
    copyAction->setStatusTip(tr("Copy the current selection's contents "
                                "to the clipboard"));
    connect(copyAction, SIGNAL(triggered()), spreadsheet, SLOT(copy()));

    pasteAction = new QAction(tr("&Paste"), this);
    pasteAction->setIcon(QIcon(":/images/paste.png"));
    pasteAction->setShortcut(QKeySequence::Paste);
    pasteAction->setStatusTip(tr("Paste the clipboard's contents into "
                                 "the current selection"));
    connect(pasteAction, SIGNAL(triggered()),
            spreadsheet, SLOT(paste()));

    deleteAction = new QAction(tr("&Delete"), this);
    deleteAction->setShortcut(QKeySequence::Delete);
    deleteAction->setStatusTip(tr("Delete the current selection's "
                                  "contents"));
    connect(deleteAction, SIGNAL(triggered()),
            spreadsheet, SLOT(del()));

    selectRowAction = new QAction(tr("&Row"), this);
    selectRowAction->setStatusTip(tr("Select all the cells in the "
                                     "current row"));
    connect(selectRowAction, SIGNAL(triggered()),
            spreadsheet, SLOT(selectCurrentRow()));

    selectColumnAction = new QAction(tr("&Column"), this);
    selectColumnAction->setStatusTip(tr("Select all the cells in the "
                                        "current column"));
    connect(selectColumnAction, SIGNAL(triggered()),
            spreadsheet, SLOT(selectCurrentColumn()));

    selectAllAction = new QAction(tr("&All"), this);
    selectAllAction->setShortcut(QKeySequence::SelectAll);
    selectAllAction->setStatusTip(tr("Select all the cells in the "
                                     "spreadsheet"));
    connect(selectAllAction, SIGNAL(triggered()),
            spreadsheet, SLOT(selectAll()));

    findAction = new QAction(tr("&Find..."), this);
    findAction->setIcon(QIcon(":/images/find.png"));
    findAction->setShortcut(QKeySequence::Find);
    findAction->setStatusTip(tr("Find a matching cell"));
    connect(findAction, SIGNAL(triggered()), this, SLOT(find()));

    goToCellAction = new QAction(tr("&Go to Cell..."), this);
    goToCellAction->setIcon(QIcon(":/images/gotocell.png"));
    goToCellAction->setShortcut(tr("Ctrl+G"));
    goToCellAction->setStatusTip(tr("Go to the specified cell"));
    connect(goToCellAction, SIGNAL(triggered()),
            this, SLOT(goToCell()));

    recalculateAction = new QAction(tr("&Recalculate"), this);
    recalculateAction->setShortcut(tr("F9"));
    recalculateAction->setStatusTip(tr("Recalculate all the "
                                       "spreadsheet's formulas"));
    connect(recalculateAction, SIGNAL(triggered()),
            spreadsheet, SLOT(recalculate()));

    sortAction = new QAction(tr("&Sort..."), this);
    sortAction->setStatusTip(tr("Sort the selected cells or all the "
                                "cells"));
    connect(sortAction, SIGNAL(triggered()), this, SLOT(sort()));

    showGridAction = new QAction(tr("&Show Grid"), this);
    showGridAction->setCheckable(true);
    showGridAction->setChecked(spreadsheet->showGrid());
    showGridAction->setStatusTip(tr("Show or hide the spreadsheet's "
                                    "grid"));
    connect(showGridAction, SIGNAL(toggled(bool)),
            spreadsheet, SLOT(setShowGrid(bool)));

    autoRecalcAction = new QAction(tr("&Auto-Recalculate"), this);
    autoRecalcAction->setCheckable(true);
    autoRecalcAction->setChecked(spreadsheet->autoRecalculate());
    autoRecalcAction->setStatusTip(tr("Switch auto-recalculation on or "
                                      "off"));
    autoRecalcAction->setShortcut(tr("Ctrl+R"));
    connect(autoRecalcAction, SIGNAL(toggled(bool)),
            spreadsheet, SLOT(setAutoRecalculate(bool)));

    aboutAction = new QAction(tr("&About..."), this);
    aboutAction->setStatusTip(tr("Show the application's About box"));
    connect(aboutAction, SIGNAL(triggered()), this, SLOT(about()));

}

void MainWindow::createMenus()
{
    fileMenu = menuBar()->addMenu(tr("&File"));
    fileMenu->addAction(newAction);
    fileMenu->addAction(openAction);
    fileMenu->addAction(saveAction);
    fileMenu->addAction(saveAsAction);
    separatorAction = fileMenu->addSeparator();
    for (int i = 0; i < MaxRecentFiles; ++i)
        fileMenu->addAction(recentFileActions[i]);
    fileMenu->addSeparator();
    fileMenu->addAction(exitAction);

    editMenu = menuBar()->addMenu(tr("&Edit"));
    editMenu->addAction(cutAction);
    editMenu->addAction(copyAction);
    editMenu->addAction(pasteAction);
    editMenu->addAction(deleteAction);

    selectSubMenu = editMenu->addMenu(tr("&Select"));
    selectSubMenu->addAction(selectRowAction);
    selectSubMenu->addAction(selectColumnAction);
    selectSubMenu->addAction(selectAllAction);

    editMenu->addSeparator();
    editMenu->addAction(findAction);
    editMenu->addAction(goToCellAction);

    optionsMenu = menuBar()->addMenu(tr("&View"));

//    optionsMenu->addAction(dock->toggleViewAction());
    optionsMenu->addAction(showResult);
    optionsMenu->addAction(showGridAction);
    optionsMenu->addAction(autoRecalcAction);

    toolsMenu = menuBar()->addMenu(tr("&Analyse"));
    toolsMenu->addAction(gInter);
    toolsMenu->addAction(statFisher);
    toolsMenu->addAction(statChi2);
    toolsMenu->addAction(cluster);
    toolsMenu->addAction(hardyW);
    toolsMenu->addAction(recalculateAction);
    toolsMenu->addAction(sortAction);

    menuBar()->addSeparator();

    helpMenu = menuBar()->addMenu(tr("&Help"));
    helpMenu->addAction(aboutAction);
//    helpMenu->addAction(aboutQtAction);
}

void MainWindow::createContextMenu()
{
    spreadsheet->addAction(cutAction);
    spreadsheet->addAction(copyAction);
    spreadsheet->addAction(pasteAction);
    spreadsheet->setContextMenuPolicy(Qt::ActionsContextMenu);
}

void MainWindow::createToolBars()
{
    fileToolBar = addToolBar(tr("&File"));
    fileToolBar->addAction(newAction);
    fileToolBar->addAction(openAction);
    fileToolBar->addAction(saveAction);

    editToolBar = addToolBar(tr("&Edit"));
    editToolBar->addAction(cutAction);
    editToolBar->addAction(copyAction);
    editToolBar->addAction(pasteAction);
    editToolBar->addSeparator();
    editToolBar->addAction(findAction);
    editToolBar->addAction(goToCellAction);
//    editToolBar->addAction(stopAction);
}

void MainWindow::createStatusBar()
{
    locationLabel = new QLabel(" W999 ");
    locationLabel->setAlignment(Qt::AlignHCenter);
    locationLabel->setMinimumSize(locationLabel->sizeHint());
    int rlen = locationLabel->sizeHint().height();

    formulaLabel = new QLabel;
    formulaLabel->setIndent(3);

    lbIndent = new QLabel;
//    int winWidth = this->width();
    lbIndent->setMinimumWidth(rlen*2);
    lbIndent->hide();

   btnStop = new QPushButton;
   btnStop->setMaximumHeight(rlen);
   btnStop->setMaximumWidth(rlen);
   btnStop->setStyleSheet(QString(
   "QPushButton {background-color: red}"));

   btnStop->setToolTip("Stop Calculating");
   btnStop->hide();

   pbCalc = new QProgressBar;
   pbCalc->setToolTip("Calculation Progress");
   pbCalc->setRange(0, 100);
   pbCalc->setFixedSize(rlen*6, rlen);
//   pbCalc->setValue(50);
   pbCalc->setTextVisible(false);
   pbCalc->setStyleSheet(QString(
   "QProgressBar {selection-background-color: rgb(128,255,128)}"));
  pbCalc->hide();

    statusBar()->addWidget(locationLabel);
//    statusBar()->addWidget(formulaLabel, 1);
    statusBar()->addWidget(formulaLabel);
    statusBar()->addWidget(lbIndent);
    statusBar()->addWidget(btnStop);
    statusBar()->addWidget(pbCalc);

    connect(spreadsheet, SIGNAL(currentCellChanged(int, int, int, int)),
            this, SLOT(updateStatusBar()));
    connect(spreadsheet, SIGNAL(modified()),
            this, SLOT(spreadsheetModified()));
   connect(btnStop, SIGNAL(clicked()), SLOT(slotStopCalc()));

    updateStatusBar();
}

void MainWindow::readSettings()
{
    QSettings settings("Moiseev I.A.", "DoctorStat");

    restoreGeometry(settings.value("geometry").toByteArray());

    recentFiles = settings.value("recentFiles").toStringList();
    updateRecentFileActions();

    bool showGrid = settings.value("showGrid", true).toBool();
    showGridAction->setChecked(showGrid);

    bool autoRecalc = settings.value("autoRecalc", true).toBool();
    autoRecalcAction->setChecked(autoRecalc);

//    bool showRes = settings.value("showResult", true).toBool();
//    dock->toggleViewAction()->setChecked(showRes);

}

void MainWindow::writeSettings()
{
    QSettings settings("Moiseev I.A.", "DoctorStat");

    settings.setValue("geometry", saveGeometry());
    settings.setValue("recentFiles", recentFiles);
    settings.setValue("showGrid", showGridAction->isChecked());
    settings.setValue("autoRecalc", autoRecalcAction->isChecked());
//    settings.setValue("showResult", showResult->isChecked());
//    bool isRes = showResult->isChecked();
//   qDebug()<<" isShowRes="<<isRes;
}

bool MainWindow::okToContinue()
{
    if (isWindowModified()) {
        int r = QMessageBox::warning(this, tr("DoctorStat"),
                        tr("The document has been modified.\n"
                           "Do you want to save your changes?"),
                        QMessageBox::Yes | QMessageBox::No
                        | QMessageBox::Cancel);
        if (r == QMessageBox::Yes) {
            return save();
        } else if (r == QMessageBox::Cancel) {
            return false;
        }
    }
    return true;
}

bool MainWindow::loadFile(const QString &fileName)
{
    if (!spreadsheet->readFile(fileName)) {
        statusBar()->showMessage(tr("Loading canceled"), 2000);
        return false;
    }

    setCurrentFile(fileName);
    statusBar()->showMessage(tr("File loaded"), 2000);
    return true;
}

bool MainWindow::saveFile(const QString &fileName)
{
    if (!spreadsheet->writeFile(fileName)) {
        statusBar()->showMessage(tr("Saving canceled"), 2000);
        return false;
    }

    setCurrentFile(fileName);
    statusBar()->showMessage(tr("File saved"), 2000);
    return true;
}

void MainWindow::setCurrentFile(const QString &fileName)
{
    curFile = fileName;
    setWindowModified(false);

    QString shownName = tr("Untitled");
    if (!curFile.isEmpty()) {
        shownName = strippedName(curFile);
        recentFiles.removeAll(curFile);
        recentFiles.prepend(curFile);
        updateRecentFileActions();
    }

    setWindowTitle(tr("%1[*] - %2").arg(shownName)
                                   .arg(tr("DoctorStat")));
}

void MainWindow::updateRecentFileActions()
{
    QMutableStringListIterator i(recentFiles);
    while (i.hasNext()) {
        if (!QFile::exists(i.next()))
            i.remove();
    }

    for (int j = 0; j < MaxRecentFiles; ++j) {
        if (j < recentFiles.count()) {
            QString text = tr("&%1 %2")
                           .arg(j + 1)
                           .arg(strippedName(recentFiles[j]));
            recentFileActions[j]->setText(text);
            recentFileActions[j]->setData(recentFiles[j]);
            recentFileActions[j]->setVisible(true);
        } else {
            recentFileActions[j]->setVisible(false);
        }
    }
    separatorAction->setVisible(!recentFiles.isEmpty());
}

QString MainWindow::strippedName(const QString &fullFileName)
{
    return QFileInfo(fullFileName).fileName();
}
// номер версии
void MainWindow::about()
{
    QMessageBox::about(this, tr("About DoctorStat"),
    tr("<p>version 1.9"
    "<p>Copyright &copy; 2013 Moiseev I.A."
    "<p>Internet: <a href=\"http://doctorstat.narod.ru\">"
         "http://doctorstat.narod.ru</a>"
    "<p>E-mail: <a href=\"mailto:imoiseev@mail.ru\">"
         "imoiseev@mail.ru</a>"
    ));
}
