#include <math.h>
#include <QDebug>
#include <QApplication> // отправить события
#include <algorithm>
#include "nfisher.h"

//// внешние переменные --------------
//extern QApplication *gpApplication;// обработка событий
//extern int gStopCalculation; // внешний флаг Останов
//extern int gProgressCalc; // прогресс (кол-во столбцов)
//int StopCalculation();  // функция внешнего останова

//==========================================
// Конструктор
// ВЫХОД:
//   err - код ошибки (успех =0)
// ВХОД:
//   maxSumTbl - максимальная сумма элементов таблицы
// ФУНКЦИЯ:
//   размещение в динамической памяти
template <typename NodeType>
base::base(int &err, int maxSumTbl){
   numCol = 0;
   err = 3;
   pMN = new NodeType [maxSumTbl];
   if(!pMN){
      return;
   }
   pDN = new NodeType [maxSumTbl];
   if(!pDN){
      return;
   }
   xMode = new int [MAXNUMCOL+1];
   if(!xMode){
      return;
   }
   lnFact = new double[maxSumTbl+1];// lnFact[]=ln(x!)
   if(!lnFact){ return;}
   lnfact(maxSumTbl);// вычисление факториалов

   pLP = new longpath(err, maxSumTbl, lnFact);
   if(err){
      QString str = "constructor longpath error=" + QString::number(err);
      qDebug()<<str;
      return;
   }
   int numc = MAXNUMCOL;
   pfe = new fexact(2, numc, lnFact);
   if(!pfe){
      return;
   }

   pPara = new para [MAXNUMCOL+1];
   if(!pPara){
      return;
   }
   xvec0 = new int [MAXNUMCOL+1];
   if(!xvec0){
      return;
   }
   col = new int [MAXNUMCOL+1];
   if(!col){
      return;
   }
   err = 0;// успех
}
//_______________________________________
// деструктор освобождает память new
base::~base(){
   delete pfe;
   delete [] lnFact;
   delete pLP;
   delete [] xMode;
   delete [] pPara;
   delete [] xvec0;
   delete [] col;
   delete [] pMN;
   delete [] pDN;
}
//========================================
// Главная функция вычисления вероятности
// ВОЗВРАТ: успех - 0
//   1, 2 - внутренние ошибки
//   10 - внешний останов
// ВХОД:
//   x1 - исходная 1-ая строка [0]
//   clmn - сумма стобцов от [0]
//   numc - кол-во столбцов
// ВЫХОД:
//   pVal - сумма вероятностей <= P0
//========================================
int base::getPvalue(int *x1, int *clmn, int numc, double &pVal){
   numCol = numc;
   for(int i=0; i<numc; i++){
      col[i] = clmn[i];
      xvec0[i] = x1[i];
   }
//   col = clmn;
//   xvec0 = x1;
   int err = initNodes();
   if(err) return 1;// ошибка
   int Nk = Ntotal;
   // k - номер столбца --------------------------------
   for(int k=numCol; k>0; k--){// уменьшается до 1
//      qDebug()<<"algoritm STEP=" << k <<"===============";
      int maxDN = 0;   // максимальное кол-во DN для к-ого шага
      int RAbsMin = 0; // минимальное R для DN's
      int numMN = linkNodes(k, Nk, maxDN, RAbsMin);
      if(k>1){
         // вычисление моды xMode[] для DN[0]
         err = getModeDN(k);
         if(err) return 2;// ошибка
      }
// вычисление Longest Path для всех дочерних узлов
      getLP( k, maxDN, Nk);
// вычисление Shortest Path для всех дочерних узлов
      if(k <= numCol-STEPSP){// пропуск первых STEPSP шагов
         iSP = 1;  // флаг вычисления SP
         getSP( k, maxDN, Nk);
      }else{
         iSP = 0;
      }
// вычисления Mother NodeS -------------------
      for(int i=0; i<numMN; i++){
         // пропустить пустые узлы
         if(pM[i].numPath==0 && k<numCol) continue;
         if(MnCalc(k, i, RAbsMin)) return 10;
      }// end MN -------------------
      Nk -= col[k-1];
      chang_pM_pD();// обмен указателями
      gProgressCalc = numCol - k + 1;// длина прогресс бара
   }// end k -------------------------
   pVal = Psum;
   return 0;// успех
}
//___________________________________________
// вычисление одной Mother Node
// ВХОД:
//   k - номер шага алгоритма (столбца)
//   iMN - индекс вычисляемой Mother Node
//   RAbsMin - (смещение) минимальное R для Daugher Nodes
// ВОЗВРАТ:
//   0 - успех
//   1 - внешний останов
//___________________________________________
int base::MnCalc(int k, int iMN, int RAbsMin){

// вычисления PDN -------------------------
   int R_PDN = pM[iMN].R - pM[iMN].XkMode;// сумма 1-ого столбца для PDN
// ri - макс.индекс (от 1) PATH, который дает вклад в p-value
   int riPDN = 0,       // ri LP для PDN
         riSP_PDN = 0;  // ri SP для PDN
   if(StopCalculation()) return 1;  // внешний останов
// расчет одного узла PDN, индекса riPDN
   getRi(k, iMN, R_PDN-RAbsMin, riPDN, riSP_PDN);

// верхнее множество дочерних узлов SU --------
   int ri = riPDN;// первый индекс (увеличивается в цикле)
   int riSP = riSP_PDN;
   int miniU = R_PDN + 1 - RAbsMin;
   int maxiU = pM[iMN].RDmax - RAbsMin;
   for(int i = miniU; i <= maxiU; i++){
//      qDebug()<<" miniU="<<miniU<<" numDN="<<numDN;
      if(StopCalculation()) return 1;
      getRi(k, iMN, i, ri, riSP);
   }

// нижнее множество дочерних узлов SL ----------
   ri = riPDN;// первый индекс (увеличивается)
   riSP = riSP_PDN;
   int maxiL = R_PDN - 1 - RAbsMin;
   int miniL = pM[iMN].RDmin - RAbsMin;
   for(int i = maxiL; i >= miniL; i--){
//      qDebug()<<" maxiL="<<maxiL<<" R_PDN="<<R_PDN;
      if(StopCalculation()) return 1;
      getRi(k, iMN, i, ri, riSP);
   }
   return 0;
}
//_____________________________________________
// вклад в p-value, вставка Path в Daughter Node
// вычисление максимального индекса ri пути PATH
// выполнения условия PAST*LP <= D*P(X0)
// ВХОД:
//    ri - индекс PATH от предыдущего шага
//    k - шаг алгоритма
//    iMN - индекс Mother Node
//    iDN - индекс Daughter Node
// ВЫХОД:
//    ri - максимальный (от 1) индекс PATH
//_____________________________________________
void base::getRi(int k, int iMN, int iDN, int &ri,
   int &rSP){
//   qDebug()<<" k="<<k<<" iMN="<<iMN
//   <<" iDN="<<iDN<<" ri="<<ri;
   int m = col[k-1];
   int n = pM[iMN].R - pD[iDN].R;
   double lk1 = pLP->ln_C(m, n);    // длина ребра
   double len = lk1 + pD[iDN].lnLP; // полный вклад
   int MnumPATH = pM[iMN].numPath;  // кол-во путей
//   qDebug()<<" MnumPATH="<<MnumPATH<<" iMN="<<iMN
//   <<" lk1="<<lk1<<" len="<<len;
   // итератор пути PATH
   list<double>::const_iterator i;

   double dd = lk1 + pD[iDN].lnPtotal - lnD0;
   i = pM[iMN].PATH.begin();
   int j;
   for(j=0; j<ri; j++){
      Psum += exp(*i + dd); // вклад в вероятность
      i++;
   }
// вычисление ri
   for(j = ri; i != pM[iMN].PATH.end(); ++i, ++j){
      if((*i + len) > lnP0) break;// выход из цикла
      Psum += exp(*i + dd);
      ri = j+1;
   }
//   if(ri>0){   // вклад в вероятность от iDN
//      for(j=0, i=pM[iMN].PATH.begin(); j<ri; i++, j++){
//         Psum += exp(*i + dd);
//      }
////      qDebug()<<" Psum="<<Psum<<" lk1="<<lk1
////      <<" pD[iDN].lnPtotal="<<pD[iDN].lnPtotal  ;
//   }

   if(k==2) return; // последний шаг алгоритма

   if(ri<MnumPATH){ // остались пути
//      qDebug()<<" ri="<<ri<<" MnumPATH="<<MnumPATH;
      double sp = pD[iDN].lnSP;// кратчаший путь
      int j=0;
      for(; i!=pM[iMN].PATH.end(); i++){// добавить ребро lk1
         double past = *i + lk1;
         if((past+sp <= lnP0) || (iSP==0)){
            // вставить пути в дочерний узел
            Mpath.push_back(past);
            j++;
         }else{
            // выбросить слишком длинные пути
//            qDebug()<<" k="<<k<<" iMN="<<iMN<<" iDN="<<iDN
//            <<" lnSP="<<sp<<" lnLP="<<pD[iDN].lnLP
//            <<" past="<<past<<" lnP0="<<lnP0;
            break;// выход из цикла
         }
      }
// вставка в дочерний узел с сохранением сортировки (по возрастанию)
      if(j){// контейнер Mpath очищается
         pD[iDN].PATH.merge(Mpath);
      }
   }
//   if(ri<MnumPATH){ // вставить пути в PATH
//      insertPath2(pD[iDN].PATH, pM[iMN].PATH, i,
//         lk1, MnumPATH-ri);
//   }
}
//____________________________________________________
// вычисление наименьшей длины Shortest Path для всех дочерних
// узлов шага k алгоритма, maxDN - максимальное кол-во DN
// Nk - сумма всех элементов дочерней таблицы
int base::getSP( int k, int maxDN, int Nk){
   int kStep = k-1; // текущий индекс столбца
   if(k>1){
      int Nk1 = Nk - col[kStep];
      int irow[2];// сумма строк
// ==========
//      int sumcol=0;
//      for(int i=0; i<kStep; i++){
//         sumcol+=col[i];
//      }
//      qDebug()<<"\n getSP kStep="<<kStep<<" Ntotal="<<Nk1
//      <<" sumcol="<<sumcol;
// ==========

      for(int i=0; i<maxDN; i++){
         // Row sums = R, Nk1-R
         irow[0] = pD[i].R;
         irow[1] = Nk1 - irow[0];
         // Column sums = col[i], i=0,..., kStep-1
         pD[i].lnSP = pfe->getSP(2, irow, kStep, col);
//         qDebug()<<" index DN="<<i<<" lnSP"<<pD[i].lnSP
//                <<" R1="<<irow[0]<<" R2="<<irow[1];
      }
   }else{// k=1 (maxDN=1)
      iSP = 0;// Shortest Path не вычислялась
   }
   return 0;// успех
}
//____________________________________________________
// вычисление наибольшей длины LP для всех дочерних
// узлов шага k алгоритма
// получить pD[i].lnLP, xMode, LP, lnPtotal для всех DN's
int base::getLP( int k, int maxDN, int Nk){
   int kStep = k-1;
   if(k>1){
      int Nk1 = Nk - col[kStep];
//      qDebug()<<" getLP kStep="<<kStep<<" N_k-1"<<Nk1;
      for(int i=0; i<maxDN; i++){
         if(i>0) setxMode(i, kStep);// получить моду xMode[]
         pD[i].lnLP = lnLength( kStep, xMode);// макс.длина
         int m = Nk1;
         int n = pD[i].R;
         pD[i].lnPtotal = pLP->ln_C(m, n);
//         qDebug()<<" index DN="<<i<<" lnLP"<<pD[i].lnLP
//               <<" lnPtotal"<<pD[i].lnPtotal;
      }
   }else{// k=1 (maxDN=1)
      pD[0].lnLP = 0;
      pD[0].lnPtotal = 0;
   }
   return 0;// успех
}
//________________________________
// класс для сортировки
class ubivorder{
public: bool operator()(const para &p, const para &q) const
   {
      return p.clmn > q.clmn;// по убыванию
   }
};
//_____________________________
// сортировать col, xvec0
void base::sortColDesc(){
   for(int i=0; i<numCol; i++){
      pPara[i].clmn = col[i];
      pPara[i].x1 = xvec0[i];
   }
//   sort(col, col+numCol);
   sort(pPara, pPara+numCol, ubivorder());
   for(int i=0; i<numCol; i++){
      col[i] = pPara[i].clmn;
      xvec0[i] = pPara[i].x1;
//      qDebug()<<" i="<<i<<" col="<<col[i]<<" x="<<xvec0[i];
   }
}
//_______________________________________
// вставить path в QList с сохранением сортировки
// numPath - размер path[]
//int insertPath(QList<double> &list, double *path,
//   int numPath);
int base::insertPath2(list<double> &lstD,
   list<double> &lstM, list<double>::const_iterator &i,
   double lk1, int numPath){
   int j1;
   list<double>::iterator j;
//   int numList = lstD.size();
   if(lstD.size()==0){// пустой список
      for( ; i != lstM.end(); i++){
         lstD.push_back(*i+lk1);
      }
//   }else if(path[numPath-1] <= list.at(0)){// вставить перед
   }else if( (lstM.back()+lk1 )<= lstD.front()){// вставить перед
      list<double>::const_reverse_iterator ir;
//      for( i=numPath-1; i>=0; i--){
      for( j1=0, ir = lstM.rbegin(); j1<numPath; ir++, j1++){
         lstD.push_front(*ir+lk1);
      }
//   }else if(path[0] >= list.at(numList-1)){// вставить после
   }else if((lstM.front()+lk1 )>= lstD.back()){// вставить после
      for( ; i != lstM.end(); i++){
//      for( i=0; i<numPath; i++){
         lstD.push_back(*i+lk1);
      }
   }else{// промежуточный вариант
      j = lstD.begin();
      if((lstM.front()+lk1 )<= lstD.front()){// вставить перед
         for( ; (*i+lk1)<= *j ; i++){
            lstD.insert(j, *i+lk1);
         }
      }// ip - индекс path[ip]>list.at(ip)
      ++j;
      while(i != lstM.end() && j != lstD.end()){
//         if(path[ip]<=list.at(it+1)){
         if((*i+lk1 )<= *j){
            lstD.insert(j, *i+lk1);
            ++i;
         }else{
            ++j;
         }
      }
// вставить после
//      for( i=ip; i<numPath; i++){
      for( ; i != lstM.end() ; i++){
//         list.append(path[i]);
         lstD.push_back(*i+lk1);
      }
   }
   return 1;// успех
}
//_______________________________________
// расчет LP для x
double base::lnLength( int k, int *x){
   double lnLP = 0;
   for(int i=0; i<k; i++){
      int m = col[i];
      int n = x[i];
      lnLP += pLP->ln_C(m, n);
   }
   return lnLP;
}
//_______________________________________
// реккурентное вычисление xMode для DN's R=1,2,...
// минимизируем выражение (x*j+1)/(Cj+1)
int base::setxMode(int iDN, int k){
//   double minExp = xMode[0];
   double minExp = (xMode[0]+1.)/(col[0]+1.);
   int Uindex=0;
   for(int i=1; i<k; i++){
      double Exp = (xMode[i]+1.)/(col[i]+1.);
      if(Exp<minExp){
         minExp = Exp;
         Uindex = i;
      }
   }
   xMode[Uindex]++;
   pD[iDN].XkMode = xMode[k-1];
//   qDebug()<<" Uindex="<<Uindex;

   return 0;// успех
}
//_______________________________________
// вычисление xMode для DN[0]
int base::getModeDN( int k){
//   int R1 = pM[0].DN[0]->R;// R1
    int R1 = pD[0].R;  // R1
   int numCl = k-1;
//   qDebug()<<" getModeDN k="<<k<<" R1="<<R1;
// вычислить моду для 1-ой DN
   int err=0;
   err = pLP->getMode(numCl, R1, col, xMode);
   if(err){
      QString str = "getModeDN longpath error=" + QString::number(err);
      qDebug()<<str;
      return 1;   // ошибка
   }
   int Xk = xMode[numCl-1];
   pD[0].XkMode = Xk;
// отладка ==================
//      int s=0;
//      for(int i=0; i<numCl; i++){
//        qDebug()<<" i="<<i<<" mode[i]="<<xMode[i];
//        s += xMode[i];
//      }
//      qDebug()<<" R1_calc="<<s<<" Xk="<<Xk;
// отладка ==================
   return 0;   // успех
}
//_______________________________________
// связать два слоя MN и DN
// возврат кол-ва MN на шаге k
// заполнить Dnode[]
int base::linkNodes(int k, int Nk, int &maxDN, int &RDAbsMin){
   int MRmin = MAX(0, R1full-Ntotal+Nk);
   int MRmax = MIN(Nk, R1full);
   int numMN = MRmax-MRmin+1;
//   qDebug()<<" MRmin="<<MRmin<<" MRmax="<<MRmax;
   RDAbsMin = R1full;
   int RDAbsMax = 0;
   for(int i=0; i<numMN; i++){// по MN
      int Rk = pM[i].R;
      int RDmin = MAX(0, Rk-col[k-1]);
      if(RDmin < RDAbsMin) RDAbsMin = RDmin;
      int RDmax = MIN(Rk, Nk-col[k-1]);
      if(RDmax > RDAbsMax) RDAbsMax = RDmax;
      int numDN = RDmax-RDmin+1;
      pM[i].numDN = numDN;
      pM[i].RDmax = RDmax;
      pM[i].RDmin = RDmin;
//      if(pM[i].numPath>0){// размер
//         pM[i].numPath = pM[i].PATH.size();
//      }
      pM[i].numPath = pM[i].PATH.size();
//      qDebug()<<" i MN="<<i<< " MR="<< Rk <<" RDmin="
//      << RDmin <<" RDmax=" << RDmax << " Nk="
//      <<Nk<<" col[k-1]="<<col[k-1]<<" numPath"
//      <<pM[i].numPath;
   }
   maxDN = RDAbsMax - RDAbsMin + 1;
   for(int j=0; j<maxDN; j++){
      pD[j].R = RDAbsMin+j;
      pD[j].PATH.clear();// очистить список
      pD[j].numPath = 0;
   }
   return numMN;  // кол-ва MN на шаге k
}
//_______________________________________
// обмен указателей M <=> D
void base::chang_pM_pD(){
   node *pTmp;
   pTmp = pM;
   pM = pD;
   pD = pTmp;
}
//_______________________________________
// вставить path в QList с сохранением сортировки
// numPath - размер path[]
int base::insertPath(QList<double> &list, double *path,
   int numPath){
   int i, it, ip;
   int numList = list.count();
   if(numList==0){// пустой список
      for( i=0; i<numPath; i++){
         list.append(path[i]);
      }
   }else if(path[numPath-1] <= list.at(0)){// вставить перед
      for( i=numPath-1; i>=0; i--){
         list.prepend(path[i]);
      }
   }else if(path[0] >= list.at(numList-1)){// вставить после
      for( i=0; i<numPath; i++){
         list.append(path[i]);
      }
   }else{// промежуточный вариант
      ip=0;// индекс path
      if(path[0] <= list.at(0)){// вставить перед
         for(ip=0; ip<numPath; ip++){
            if(path[ip] <= list.at(ip)){
               list.insert(ip, path[ip]);
            }else{
               break;
            }
         }
      }// ip - индекс path[ip]>list.at(ip)
      it = ip; // индекс list
      numList = list.count();
      while(ip<numPath && it<numList-1){
         if(path[ip]<=list.at(it+1)){
            list.insert(it+1, path[ip]);
            numList++;
            ip++;
         }
         it++;
//         qDebug()<<" it="<<it;
      }
      // вставить после
      for( i=ip; i<numPath; i++){
         list.append(path[i]);
      }
   }
   return 1;// успех
}
//______________________________
// Вычисление логарифма факториала *lnFact
// вызов функции натурального логарифма log()
// только для нечетного аргумента
int base::lnfact(int maxFact){
   lnFact[0] = 0;
   lnFact[1] = 0;
   if(maxFact >= 2) lnFact[2] = log(2.);
   for(int i=3; i<=maxFact; i += 2){
      lnFact[i] = lnFact[i - 1] + log(i);
      int j = i + 1;
      if (j <= maxFact){
         lnFact[j] = lnFact[i] + lnFact[2] +
         lnFact[j / 2] - lnFact[j / 2 - 1];
      }
//      qDebug()<<" i="<<i<<" lnFact[i]="<<lnFact[i];
   }
   return 0;
}

// производный класс ===================================
// конструктор точного расчета
nfisher::nfisher(int &err, int maxSumTbl)
   : base (err, maxSumTbl)
{
   if(err) return;   // ошибка
//   pMN = new node [maxSumTbl];
//   if(!pMN){
//      return;
//   }
//   pDN = new node [maxSumTbl];
//   if(!pDN){
//      return;
//   }
   err = 0;// успех
}
//_______________________________________
// деструктор освобождает память new
nfisher::~nfisher(){
//   delete [] pMN;
//   delete [] pDN;
}
//_______________________________________
// инициализация MN and DN
int nfisher::initNodes(){
   R1full = 0;
   Ntotal = 0;
   // сортировать col, xvec0 в порядке убывания
   sortColDesc();
   for(int i=0; i<numCol; i++){
      Ntotal += col[i];
      R1full += xvec0[i];
   }
   pM = pMN;
   pD = pDN;
   pM[0].R = R1full;
   pM[0].PATH.clear();
//   pM[0].PATH << 0;
   pM[0].PATH.push_back(0);

   pM[0].numPath = 1;
// вычислить моду xMode
   int err=0;
   err = pLP->getMode(numCol, R1full, col, xMode);
   if(err){
      QString str = "initNodes longpath error=" + QString::number(err);
      qDebug()<<str;
      return -1;
   }
// отладка ===========
//   for(int i=0; i<numCol;i++){
//      qDebug()<<" i="<<i<<" xMode="<<xMode[i];
//   }
// отладка ===========
   pM[0].XkMode = xMode[numCol-1];
   // вычислить ln(Longest Pass)
   lnP0 = lnLength(numCol, xvec0);

//   double pg = EPSF;
//   if(lnP0<0) pg = -pg;
//   lnP0 *= 1 + pg;
   lnP0 += EPSF;

   lnD0 = pLP->ln_C(Ntotal, R1full);
   D0 = exp(lnD0);
//   qDebug()<<" lnP0="<<lnP0<<" lnD0="<<lnD0<<" D0="<<D0;
   Psum = 0;
   return 0;
}



