//==================================
// точный тест Фишера для таблицы сопряженности 2хN
// Сетевой алгоритм по статье:
// Requena, Martin, 2005 "A major improvement to the network..."
//==================================

#ifndef NFISHER_H
#define NFISHER_H

#include <cstdlib>
#include <list>
#include <algorithm>
using namespace std;

#include <math.h>
#include <QDebug>
#include <QApplication> // отправить события

#include "longpath.h"
#include "fexact2.h"
#include "memfunc.h"

// внешние переменные --------------
extern QApplication *gpApplication;// обработка событий
extern int gStopCalculation; // внешний флаг Останов
extern int gProgressCalc; // прогресс (кол-во столбцов)
int StopCalculation();// останов вычислений
//extern int gNewError;

// контейнер сортировки столбцов таблицы 2xN
struct para{
   int clmn;// столбец
   int x1;  // строка 1
};
//----------------------------
// пройденный путь PATH описывается 2-мя типами контейнеров
//   1 - double - точное вычисление вероятности
//   2 - struct path приближенное ...
struct path{
   int dlina; // длина пути
//   int freq;  // частота встречаемости (1,2,...)
   double freq;  // частота встречаемости (1,2,...)
};
//----------------------------
// шаблон узла, переменная типа пути Node
// Node = double или path
template <class Node>
struct node{
   list<Node> PATH;// логарифм длины пройденных путей
   int numPath;// кол-во PATH-ов (размер контейнера)
   int R;      // сумма строки
   int RDmax;  // максимальное R для дочерних узлов DN
   int RDmin;  // минимальное R для дочерних узлов DN
   int XkMode; // x k-ое моды (longest path)
   double lnLP;   // логарифм longest path для 0-node
   double lnSP;   // логарифм shortest path для 0-node
   double lnPtotal;// логарифм вклада DN в суммарную вероятность
};
//===================================================
// базовый класс вычисления p-value Фишера
//   Node - шаблон (переменный тип) узла сети
template <class Node>
class basef{
public:
   basef(int &err, int maxSumTbl);// конструктор
   ~basef();       // деструктор
   int getPvalue(int *x, int *col, int numc, double &pVal);
protected:
   // КОНСТАНТЫ
   static const int STEPSP = 3;// кол-во пропуска первых шагов SP
//   static const int STEPSP = 1;// кол-во пропуска первых шагов SP
   static const double EPSF = 1e-10; // погрешность вычислений
   static const int MINRAM = 10;// мин.кол-во свободной физ.памяти, МБ
   // ПЕ ЕМЕННЫЕ

//   int deg10;  // степень 10 для умножения ln
   double deg10;  // степень 10 для умножения ln
   int iSP;    // флаг расчета SP - shortest path
   int numCol; // кол-во столбцов
   int R1full; // сумма 1-ой строки
   int Ntotal; // сумма всех элементов
   int *col;   // сумма стобцов от [0]
   int *xvec0; // исходная 1-ая строка [0]
   int *xMode; // строка моды от [0]

   node<Node> *pMN;  // mother nodes
   node<Node> *pDN;  // daughter nodes
   node<Node> *pM;
   node<Node> *pD;
   list<Node> Mpath;// предыдущий путь + длина ребра
   longpath *pLP; // класс расчета моды (максимального пути)
   fexact *pfe;   // класс расчета минимального пути
   double *lnFact;// ln(x!)
   double lnP0;   // логарифм исходного пути
   double lnD0;   // логарифм D = (N, R1)
   double D0;     // D = (N, R1)
   double Psum;   // сумма вероятностей = p-value
   para *pPara;
   // ФУНКЦИИ
   virtual void virtInit()=0;
   virtual int getRi(int k, int iMN, int iDN, int &ri0) = 0;

   int initNodes();   // виртуальная функция
   int lnfact(int maxFact);
   int linkNodes(int k, int Nk, int &maxDN, int &RDAbsMin);
   void chang_pM_pD();
   int getLP( int k, int maxDN, int Nk);
   int getSP( int k, int maxDN, int Nk);
   int getModeDN( int k);
   int setxMode(int i, int k);
   double lnLength( int k, int *x);
   int MnCalc(int k, int iMN, int RAbsMin);
   void sortColDesc();
};
//_______________________________________
// Конструктор ТОЧНОГО расчета
// ВЫХОД:
//   err - код ошибки (успех =0)
// ВХОД:
//   maxSumTbl - максимальная сумма элементов таблицы
// ФУНКЦИЯ:
//   размещение в памяти new
template <class Node>
basef<Node>::basef(int &err, int maxSumTbl){
   deg10 = 1;
   numCol = 0;
   err = 3;
   pMN = new node<Node> [maxSumTbl];
   if(!pMN){
      return;
   }
   pDN = new node<Node> [maxSumTbl];
   if(!pDN){
      return;
   }
   xMode = new int [MAXNUMCOL+1];
   if(!xMode){
      return;
   }
   lnFact = new double[maxSumTbl+1];// lnFact[]=ln(x!)
   if(!lnFact){ return;}
   lnfact(maxSumTbl);// вычисление факториалов

   pLP = new longpath(err, maxSumTbl, lnFact);
   if(err){
      QString str = "constructor longpath error=" + QString::number(err);
      qDebug()<<str;
      return;
   }
   int numc = MAXNUMCOL;
   pfe = new fexact(2, numc, lnFact);
   if(!pfe){
      return;
   }

   pPara = new para [MAXNUMCOL+1];
   if(!pPara){
      return;
   }
   xvec0 = new int [MAXNUMCOL+1];
   if(!xvec0){
      return;
   }
   col = new int [MAXNUMCOL+1];
   if(!col){
      return;
   }
   err = 0;// успех
}
//___________________________
// деструктор базового класса
// освобождает память new
template <class Node>
basef<Node>::~basef(){
   delete pfe;
   delete [] lnFact;
   delete pLP;
   delete [] xMode;
   delete [] pPara;
   delete [] xvec0;
   delete [] col;
   delete [] pMN;
   delete [] pDN;
}
//========================================
// Главная функция вычисления вероятности pVal таблицы 2xN
// ВОЗВ АТ: успех - 0
//   1, 2 - внутренние ошибки
//   10 - внешний останов
// ВХОД:
//   x1 - исходная 1-ая строка от [0]
//   clmn - сумма стобцов от [0]
//   numc - кол-во столбцов
// ВЫХОД:
//   pVal - сумма вероятностей <= P0 - вероятность начальной таблицы
//========================================
template <class Node>
int basef<Node>::getPvalue(int *x1, int *clmn, int numc,
double &pVal){
   numCol = numc;
   for(int i=0; i<numc; i++){
      col[i] = clmn[i];
      xvec0[i] = x1[i];
   }
   int err = initNodes();
   if(err) return 1;// ошибка
   int Nk = Ntotal;// сумма элементов подтаблицы (без первых столбцов)
   // k - номер столбца --------------------------------
   for(int k=numCol; k>0; k--){// уменьшается до 1
//      qDebug()<<"algoritm STEP=" << k <<"===============";
      int maxDN = 0;   // максимальное кол-во DN для к-ого шага
      int RAbsMin = 0; // минимальное R для DN's
      // связать два слоя MN и DN
      int numMN = linkNodes(k, Nk, maxDN, RAbsMin);
      if(k>1){// для всех шагов, кроме последнего
         // вычисление моды xMode[] для DN[0]
         err = getModeDN(k);
         if(err) return 2;// ошибка
      }
      // вычисление Longest Path для всех дочерних узлов
      getLP( k, maxDN, Nk);
      // вычисление Shortest Path для всех дочерних узлов
      if(k <= numCol-STEPSP){// пропуск первых STEPSP шагов
         iSP = 1;  // флаг вычисления SP
         getSP( k, maxDN, Nk);
      }else{
         iSP = 0;// короткий путь SP не вычислялся
      }
      // вычисления Mother NodeS -------------------
      for(int i=0; i<numMN; i++){
         // пропустить пустые узлы
         if(pM[i].numPath==0 && k<numCol) continue;
         // вычисление одной Mother Node
         err = MnCalc(k, i, RAbsMin);
         if(err) return (err+9);
//         if(MnCalc(k, i, RAbsMin)) return 10;
      }// end MN -------------------
      Nk -= col[k-1];
      chang_pM_pD();// обмен указателями
      gProgressCalc = numCol - k + 1;// длина полосы прогресса
   }// end k -------------------------
   pVal = Psum;
   return 0;// успех
}
//_______________________________________
// инициализация MN and DN
template <class Node>
int basef<Node>::initNodes(){
	R1full = 0;
	Ntotal = 0;
	// сортировать col, xvec0 в порядке убывания
	sortColDesc();
	for(int i=0; i<numCol; i++){
		if(xvec0[i]>col[i]){
			qDebug()<<"Error: column sum < row value";
			return -1;
		}
		Ntotal += col[i];
		R1full += xvec0[i];
	}
	pM = pMN;
	pD = pDN;
	pM[0].R = R1full;
	pM[0].PATH.clear();
//	pM[0].PATH.push_back(0);// множитель =1
	pM[0].numPath = 1;
// вычислить моду xMode
	int err=0;
	err = pLP->getMode(numCol, R1full, col, xMode);
	if(err){
		QString str = "initNodes longpath error=" + QString::number(err);
		qDebug()<<str;
		return -1;
	}
// отладка ===========
//   for(int i=0; i<numCol;i++){
//      qDebug()<<" i="<<i<<" xMode="<<xMode[i];
//   }
// отладка ===========
	pM[0].XkMode = xMode[numCol-1];
	// вычислить ln(исходного пути) без знаменателя lnD0
	lnP0 = this->lnLength(numCol, xvec0);
//	lnP0 *= deg10;
	virtInit();// начальная длина пути
	lnP0 += EPSF;

	lnD0 = pLP->ln_C(Ntotal, R1full);
	D0 = exp(lnD0);
//   qDebug()<<" lnP0="<<lnP0<<" lnD0="<<lnD0<<" D0="<<D0;
	Psum = 0;
	return 0;
}
//___________________________________________
// вычисление одной Mother Node
// ВХОД:
//   k - номер шага алгоритма (столбца)
//   iMN - индекс вычисляемой Mother Node
//   RAbsMin - (смещение) минимальное R для Daugher Nodes
// ВОЗВ АТ:
//   0 - успех
//   1 - внешний останов
//   2 - нехватка памяти
template <class Node>
int basef<Node>::MnCalc(int k, int iMN, int RAbsMin){
	// вычисления PDN (principal daughter node)
	int R_PDN = pM[iMN].R - pM[iMN].XkMode;// сумма 1-ого столбца для PDN
	// ri - макс.индекс (от 1) PATH, который дает вклад в p-value
	int riPDN = 0;       // ri LP для PDN
//			riSP_PDN = 0;  // отладка (не используется)
	if(StopCalculation()) return 1;  // внешний останов

	int memav = getAvailRAM();
	if(memav>0 && memav<MINRAM){
		return 2;// нехватка памяти
	}

	// расчет одного узла PDN, индекса riPDN
	int RDater0 = R_PDN-RAbsMin;
	int err = getRi(k, iMN, RDater0, riPDN);
	if(err){
		return err;
	}

	// верхнее подмножество узлов SU (subsuccession up)
	int ri = riPDN;// первый индекс (увеличивается в цикле)
//	int riSP = riSP_PDN;
	int miniU = R_PDN + 1 - RAbsMin;
	int maxiU = pM[iMN].RDmax - RAbsMin;
	for(int i = miniU; i <= maxiU; i++){
//      qDebug()<<" miniU="<<miniU<<" numDN="<<numDN;
		if(StopCalculation()) return 1;
		err = getRi(k, iMN, i, ri);
		if(err){
			return err;
		}
	}
	// нижнее подмножество узлов SL (subsuccession low)
	ri = riPDN;// первый индекс (уменьшается)
//	riSP = riSP_PDN;
	int maxiL = R_PDN - 1 - RAbsMin;
	int miniL = pM[iMN].RDmin - RAbsMin;
	for(int i = maxiL; i >= miniL; i--){
//      qDebug()<<" maxiL="<<maxiL<<" R_PDN="<<R_PDN;
		if(StopCalculation()) return 1;
		err = getRi(k, iMN, i, ri);
		if(err){
			return err;
		}
	}
	return 0;
}
//____________________________________________________
// вычисление наименьшей длины Shortest Path для всех дочерних
// узлов шага k алгоритма, maxDN - максимальное кол-во DN
// Nk - сумма всех элементов дочерней таблицы
template <class Node>
int basef<Node>::getSP( int k, int maxDN, int Nk){
   int kStep = k-1; // текущий индекс столбца
   if(k>1){
      int Nk1 = Nk - col[kStep];
      int irow[2];// сумма строк
// ==========
//      int sumcol=0;
//      for(int i=0; i<kStep; i++){
//         sumcol+=col[i];
//      }
//      qDebug()<<"\n getSP kStep="<<kStep<<" Ntotal="<<Nk1
//      <<" sumcol="<<sumcol;
// ==========

      for(int i=0; i<maxDN; i++){
         // Row sums = R, Nk1-R
         irow[0] = pD[i].R;
         irow[1] = Nk1 - irow[0];
         // Column sums = col[i], i=0,..., kStep-1
         pD[i].lnSP = pfe->getSP(2, irow, kStep, col);
//         qDebug()<<" index DN="<<i<<" lnSP"<<pD[i].lnSP
//                <<" R1="<<irow[0]<<" R2="<<irow[1];
      }
   }else{// k=1 (maxDN=1)
      iSP = 0;// Shortest Path не вычислялась
   }
   return 0;// успех
}
//____________________________________________________
// вычисление наибольшей длины LP для всех дочерних
// узлов шага k алгоритма
// получить pD[i].lnLP, xMode, LP, lnPtotal для всех DN's
template <class Node>
int basef<Node>::getLP( int k, int maxDN, int Nk){
   int kStep = k-1;
   if(k>1){
      int Nk1 = Nk - col[kStep];
//      qDebug()<<" getLP kStep="<<kStep<<" N_k-1"<<Nk1;
      for(int i=0; i<maxDN; i++){
         if(i>0) setxMode(i, kStep);// получить моду xMode[]
         pD[i].lnLP = lnLength( kStep, xMode);// макс.длина
         int m = Nk1;
         int n = pD[i].R;
         pD[i].lnPtotal = pLP->ln_C(m, n);
//         qDebug()<<" index DN="<<i<<" lnLP"<<pD[i].lnLP
//               <<" lnPtotal"<<pD[i].lnPtotal;
      }
   }else{// k=1 (maxDN=1)
      pD[0].lnLP = 0;
      pD[0].lnPtotal = 0;
   }
   return 0;// успех
}
//________________________________
// класс сравнения структур для сортировки STL
class ubivorder{
public: bool operator()(const para &p, const para &q) const
   {
      return p.clmn > q.clmn;// по убыванию
   }
};
//_____________________________
// сортировать col, xvec0
template <class Node>
void basef<Node>::sortColDesc(){
   for(int i=0; i<numCol; i++){
      pPara[i].clmn = col[i];
      pPara[i].x1 = xvec0[i];
   }
//   sort(col, col+numCol);
   sort(pPara, pPara+numCol, ubivorder());
   for(int i=0; i<numCol; i++){
      col[i] = pPara[i].clmn;
      xvec0[i] = pPara[i].x1;
//      qDebug()<<" i="<<i<<" col="<<col[i]<<" x="<<xvec0[i];
   }
}
//_______________________________________
// расчет LP для x
template <class Node>
double basef<Node>::lnLength( int k, int *x){
   double lnLP = 0;
   for(int i=0; i<k; i++){
      int m = col[i];
      int n = x[i];
      lnLP += pLP->ln_C(m, n);
   }
   return lnLP;
}
//_______________________________________
// реккурентное вычисление xMode для DN's R=1,2,...
// минимизируем выражение (x*j+1)/(Cj+1)
template <class Node>
int basef<Node>::setxMode(int iDN, int k){
//   double minExp = xMode[0];
   double minExp = (xMode[0]+1.)/(col[0]+1.);
   int Uindex=0;
   for(int i=1; i<k; i++){
      double Exp = (xMode[i]+1.)/(col[i]+1.);
      if(Exp<minExp){
         minExp = Exp;
         Uindex = i;
      }
   }
   xMode[Uindex]++;
   pD[iDN].XkMode = xMode[k-1];
//   qDebug()<<" Uindex="<<Uindex;

   return 0;// успех
}
//_______________________________________
// вычисление xMode для DN[0]
template <class Node>
int basef<Node>::getModeDN( int k){
//   int R1 = pM[0].DN[0]->R;// R1
    int R1 = pD[0].R;  // R1
   int numCl = k-1;
//   qDebug()<<" getModeDN k="<<k<<" R1="<<R1;
// вычислить моду для 1-ой DN
   int err=0;
   err = pLP->getMode(numCl, R1, col, xMode);
   if(err){
      QString str = "getModeDN longpath error=" + QString::number(err);
      qDebug()<<str;
      return 1;   // ошибка
   }
   int Xk = xMode[numCl-1];
   pD[0].XkMode = Xk;
// отладка ==================
//      int s=0;
//      for(int i=0; i<numCl; i++){
//        qDebug()<<" i="<<i<<" mode[i]="<<xMode[i];
//        s += xMode[i];
//      }
//      qDebug()<<" R1_calc="<<s<<" Xk="<<Xk;
// отладка ==================
   return 0;   // успех
}
//_______________________________________
// связать два слоя MN и DN
// возврат кол-ва MN на шаге k
// заполнить Dnode[]
template <class Node>
int basef<Node>::linkNodes(int k, int Nk, int &maxDN,
int &RDAbsMin){
   int MRmin = MAX(0, R1full-Ntotal+Nk);
   int MRmax = MIN(Nk, R1full);
   int numMN = MRmax-MRmin+1;
//   qDebug()<<" MRmin="<<MRmin<<" MRmax="<<MRmax;
   RDAbsMin = R1full;
   int RDAbsMax = 0;
   for(int i=0; i<numMN; i++){// по MN
      int Rk = pM[i].R;
      int RDmin = MAX(0, Rk-col[k-1]);
      if(RDmin < RDAbsMin) RDAbsMin = RDmin;
      int RDmax = MIN(Rk, Nk-col[k-1]);
      if(RDmax > RDAbsMax) RDAbsMax = RDmax;
//      int numDN = RDmax-RDmin+1;
//      pM[i].numDN = numDN;
      pM[i].RDmax = RDmax;
      pM[i].RDmin = RDmin;
//      if(pM[i].numPath>0){// размер
//         pM[i].numPath = pM[i].PATH.size();
//      }
      pM[i].numPath = pM[i].PATH.size();
//      qDebug()<<" i MN="<<i<< " MR="<< Rk <<" RDmin="
//      << RDmin <<" RDmax=" << RDmax << " Nk="
//      <<Nk<<" col[k-1]="<<col[k-1]<<" numPath"
//      <<pM[i].numPath;
   }
   maxDN = RDAbsMax - RDAbsMin + 1;
   for(int j=0; j<maxDN; j++){
      pD[j].R = RDAbsMin+j;
      pD[j].PATH.clear();// очистить список
      pD[j].numPath = 0;
   }
   return numMN;  // кол-ва MN на шаге k
}
//_______________________________________
// обмен указателей M <=> D
template <class Node>
void basef<Node>::chang_pM_pD(){
   node<Node> *pTmp;
   pTmp = pM;
   pM = pD;
   pD = pTmp;
}
//_______________________________________
// вставить path в QList с сохранением сортировки
// numPath - размер path[]
//__________________________________________
// Вычисление логарифма факториала *lnFact
// вызов функции натурального логарифма log()
// только для нечетного аргумента (для ускорения)
template <class Node>
int basef<Node>::lnfact(int maxFact){
   lnFact[0] = 0;
   lnFact[1] = 0;
   if(maxFact >= 2) lnFact[2] = log(2.);
   for(int i=3; i<=maxFact; i += 2){
      lnFact[i] = lnFact[i - 1] + log(i);
      int j = i + 1;
      if (j <= maxFact){
         lnFact[j] = lnFact[i] + lnFact[2] +
         lnFact[j / 2] - lnFact[j / 2 - 1];
      }
//      qDebug()<<" i="<<i<<" lnFact[i]="<<lnFact[i];
   }
   return 0;
}
//===================================================
// производный класс ТОЧНОГО вычисления P-Val Фишера
template <class Node>
class dfex : public basef <Node> {
public:
// переменные базового класса
   using basef<Node>:: deg10;  // тип int множитель ln
   using basef<Node>:: EPSF; // погрешность вычислений
   using basef<Node>:: iSP;    // флаг расчета SP - shortest path
   using basef<Node>:: numCol; // кол-во столбцов
   using basef<Node>:: R1full; // сумма 1-ой строки
   using basef<Node>:: Ntotal; // сумма всех элементов
   using basef<Node>:: col;   // сумма стобцов от [0]
   using basef<Node>:: xvec0; // исходная 1-ая строка [0]
   using basef<Node>:: xMode; // строка моды от [0]
   using basef<Node>:: pMN;  // mother nodes
   using basef<Node>:: pDN;  // daughter nodes
   using basef<Node>:: pM;
   using basef<Node>:: pD;
   using basef<Node>:: Mpath;// предыдущий путь + длина ребра
   using basef<Node>:: pLP; // класс расчета моды (максимального пути)
   using basef<Node>:: pfe;   // класс расчета минимального пути
   using basef<Node>:: lnFact;// ln(x!)
   using basef<Node>:: lnP0;   // логарифм исходного пути
   using basef<Node>:: lnD0;   // логарифм D = (N, R1)
   using basef<Node>:: D0;     // D = (N, R1)
   using basef<Node>:: Psum;   // сумма вероятностей = p-value
   using basef<Node>:: pPara;
   using basef<Node>:: MINRAM;

   dfex(int &err, int maxSumTbl);// конструктор
   ~dfex();       // деструктор
protected:
   virtual void virtInit();
   virtual int getRi(int k, int iMN, int iDN, int &ri0);
};
//______________________________
// конструктор ТОЧНОГО Фишера
//   err - код ошибки (успех =0)
//   maxSumTbl - максимальная сумма элементов таблицы
//   numd - точность (=1,10,100,...) для умножения ln
template <class Node>
dfex<Node>::dfex(int &err, int maxSumTbl)
   : basef<Node> (err, maxSumTbl){
   if(err) return;   // ошибка
   err = 0;// успех
}
//_______________________________________
// деструктор освобождает память new
template <class Node>
dfex<Node>::~dfex(){// пустой, т.к. память не выделяется
}
//____________________
// инициализация начального пути
template <class Node>
void dfex<Node>::virtInit(){
   pM[0].PATH.push_back(0);// множитель =1
}
//_____________________________________________
// вычисление вклада в p-value для одной MN и одной DN
// вставка оставшихся длинных путей (плюс длина ребра) из MN в DN
// вычисление максимального индекса ri пути PATH для MN
// удовлетворяющему условию для логарифмов: PAST(ri)+LP <= lnP0 (1)
// ВХОД:
//    ri - индекс пути PATH от предыдущего DN
//    k - шаг алгоритма (номер столбца)
//    iMN - индекс Mother Node
//    iDN - индекс Daughter Node
// ВЫХОД:
//    ri - максимальный (от 1) индекс PATH, удовл.условию (1)
template <class Node>
int dfex<Node>::getRi(int k, int iMN, int iDN, int &ri){
//   qDebug()<<" k="<<k<<" iMN="<<iMN
//   <<" iDN="<<iDN<<" ri="<<ri;
   int m = col[k-1];
   int n = pM[iMN].R - pD[iDN].R;
   // длина ребра от iMN до iDN
   double lk1 = pLP->ln_C(m, n);
   // полный вклад в вероятность от iDN + длина ребра
   double len = lk1 + pD[iDN].lnLP;
   // кол-во предыдущих путей у iMN
   int MnumPATH = pM[iMN].numPath;
//   qDebug()<<" MnumPATH="<<MnumPATH<<" iMN="<<iMN
//   <<" lk1="<<lk1<<" len="<<len;
   // итератор пути PATH
   list<double>::const_iterator i;

// вклад от путей с индексом от 0 до ri ===========
   double dd = lk1 + pD[iDN].lnPtotal - lnD0;
   i = pM[iMN].PATH.begin();
   int j;
   for(j=0; j<ri; j++){
      Psum += exp(*i + dd); // вклад в вероятность
      i++;
   }
// вычисление ri ================================
   for(j = ri; i != pM[iMN].PATH.end(); ++i, ++j){
      if((*i + len) > lnP0) break;// выход из цикла
      Psum += exp(*i + dd);
      ri = j+1;
   }

   if(k==2) return 0; // последний шаг алгоритма

   // вставка путей в промежуточный контейнер ============
   if(ri<MnumPATH){ // остались пути
//      int memav = getAvailRAM();
//      if(memav>0 && memav<MINRAM){
//         return 2;// нехватка памяти
//      }

//      try{
         //      qDebug()<<" ri="<<ri<<" MnumPATH="<<MnumPATH;
         double sp = pD[iDN].lnSP;// кратчайший путь
         int j=0;
         for(; i!=pM[iMN].PATH.end(); i++){// добавить ребро lk1
            double past = *i + lk1;
            if((past+sp <= lnP0) || (iSP==0)){
               // вставить пути в дочерний узел
               Mpath.push_back(past);
               j++;
            }else{// удалить слишком длинные пути из расчета
               //            qDebug()<<" k="<<k<<" iMN="<<iMN<<" iDN="<<iDN
               //            <<" lnSP="<<sp<<" lnLP="<<pD[iDN].lnLP
               //            <<" past="<<past<<" lnP0="<<lnP0;
               break;// выход из цикла
            }
         }
         // перемещение контейнера в DN с сохранением сортировки по возрастанию
         if(j){// контейнер Mpath очищается !!!
            pD[iDN].PATH.merge(Mpath);
         }
//      }catch(std::bad_alloc& ba){// нехватка памяти
//      }catch(...){// нехватка памяти
//         return 2;
//      }
   }
   return 0;// успех
}
//===================================================
// производный класс П ИБЛИЖЕННОГО вычисления P-Val Фишера
template <class Node>
class dfish : public basef <Node> {
public:
// переменные базового класса
   using basef<Node>:: deg10;  // тип int множитель ln
   using basef<Node>:: EPSF; // погрешность вычислений
   using basef<Node>:: iSP;    // флаг расчета SP - shortest path
   using basef<Node>:: numCol; // кол-во столбцов
   using basef<Node>:: R1full; // сумма 1-ой строки
   using basef<Node>:: Ntotal; // сумма всех элементов
   using basef<Node>:: col;   // сумма стобцов от [0]
   using basef<Node>:: xvec0; // исходная 1-ая строка [0]
   using basef<Node>:: xMode; // строка моды от [0]
   using basef<Node>:: pMN;  // mother nodes
   using basef<Node>:: pDN;  // daughter nodes
   using basef<Node>:: pM;
   using basef<Node>:: pD;
   using basef<Node>:: Mpath;// предыдущий путь + длина ребра
   using basef<Node>:: pLP; // класс расчета моды (максимального пути)
   using basef<Node>:: pfe;   // класс расчета минимального пути
   using basef<Node>:: lnFact;// ln(x!)
   using basef<Node>:: lnP0;   // логарифм исходного пути
   using basef<Node>:: lnD0;   // логарифм D = (N, R1)
   using basef<Node>:: D0;     // D = (N, R1)
   using basef<Node>:: Psum;   // сумма вероятностей = p-value
   using basef<Node>:: pPara;
   using basef<Node>:: MINRAM;
//   list<Node>::const_iterator iter;
//
   dfish(int &err, int maxSumTbl, int numd);// конструктор
   ~dfish();       // деструктор
protected:
   double multy;// множитель логарифмов
   virtual void virtInit();
//   void virtInit();
   virtual int getRi(int k, int iMN, int iDN, int &ri0);
//   void movePath(list<path> &lstD, list<path> &lstM);
private:

};
//______________________________
// конструктор П ИБЛИЖЕННОГО Фишера
//   err - код ошибки (успех =0)
//   maxSumTbl - максимальная сумма элементов таблицы
//   numd - степень 10 (0<=numd<=9) для умножения ln
template <class Node>
dfish<Node>::dfish(int &err, int maxSumTbl, int numd)
   : basef<Node> (err, maxSumTbl){
   if(err) return;   // ошибка
   if(numd<0 || numd>9){
      err=1;
      return;
   }
   int st10=1;
   for(int i=0; i<numd; i++){
      st10 *= 10;
   }
   deg10 = st10;
   multy = 1./st10;
   err = 0;// успех
}
//_______________________________________
// деструктор освобождает память new
template <class Node>
dfish<Node>::~dfish(){// пустой, т.к. память не выделяется
}
//______________________________________
// инициализация начального пути
template <class Node>
void dfish<Node>::virtInit(){
   Node tmp1;
   tmp1.dlina = 0;
   tmp1.freq = 1;
   pM[0].PATH.push_back(tmp1);// множитель =1
}
//_____________________________________________
// вычисление вклада в p-value для одной MN и одной DN
// вставка оставшихся длинных путей (плюс длина ребра) из MN в DN
// вычисление максимального индекса ri пути PATH для MN
// удовлетворяющему условию для логарифмов: PAST(ri)+LP <= lnP0 (1)
// ВХОД:
//    ri - индекс пути PATH от предыдущего DN
//    k - шаг алгоритма (номер столбца)
//    iMN - индекс Mother Node
//    iDN - индекс Daughter Node
// ВЫХОД:
//    ri - максимальный (от 1) индекс PATH, удовл.условию (1)
template <class Node>
      int dfish<Node>::getRi(int k, int iMN, int iDN, int &ri){
   //   qDebug()<<" k="<<k<<" iMN="<<iMN
   //   <<" iDN="<<iDN<<" ri="<<ri;
   int m = col[k-1];
   int n = pM[iMN].R - pD[iDN].R;
   // длина ребра от iMN до iDN
   double lk1 = pLP->ln_C(m, n);
   //   qDebug()<<" MnumPATH="<<MnumPATH<<" iMN="<<iMN
   //   <<" lk1="<<lk1<<" len="<<len;
   // итератор пути PATH
   list<path>::const_iterator iter;

   // вклад от путей с индексом от 0 до ri ===========
   // вклад в вероятность от DN
   double dd = lk1 + pD[iDN].lnPtotal - lnD0;
   double lb; // длина путей до MN
   iter = pM[iMN].PATH.begin();
   int j;// индекс пути
   for(j=0; j<ri; j++){
      lb = (*iter).dlina * multy;
      double num = (*iter).freq;
      Psum += num * exp(lb + dd); // вклад в вероятность
      iter++;
   }

   // вычисление ri ================================
   // максимальная длина второй части пути
   double len = lk1 + pD[iDN].lnLP;
   for(j = ri; iter != pM[iMN].PATH.end(); ++iter, ++j){
      lb = (*iter).dlina * multy;
      if((lb + len) > lnP0) break;// выход из цикла
      double num = (*iter).freq;
      Psum += num * exp(lb + dd); // вклад в вероятность
      ri = j+1;
   }

   if(k==2) return 0; // предпоследний шаг алгоритма

   // вставка путей в дочерний узел ============
   // кол-во предыдущих путей у iMN
   int MnumPATH = pM[iMN].numPath;
   if(ri<MnumPATH){ // остались необработанные пути
//      int memav = getAvailRAM();
//      if(memav>0 && memav<MINRAM){
//         return 2;// нехватка памяти
//      }
//      try{
         //      qDebug()<<" ri="<<ri<<" MnumPATH="<<MnumPATH;
         double sp = pD[iDN].lnSP;// кратчайший путь
         j=0;
         list<path>::iterator jt = pD[iDN].PATH.begin();
         for(; iter != pM[iMN].PATH.end(); iter++){
            // добавить ребро lk1
            double past = (*iter).dlina * multy + lk1;
            if((past+sp <= lnP0) || (iSP==0)){
               path tmp;// временный контейнер
               // вставить пути в дочерний узел
               double ost = lk1*deg10;
               if(ost<0){
                  qDebug()<<" dlina linka<0";
                  exit(1);
               }
               int itmp = ost;
               if(itmp<0){
                  return 3;
               }
               //               ost -= itmp;
               //               if(ost > 0.5){
               //                  itmp++;
               //               }
               //               int dlin = (*iter).dlina + lk1*deg10;
               int dlin = (*iter).dlina + itmp;
               if(dlin<0){
                  return 3;
               }
               if(jt == pD[iDN].PATH.end()){// вставить после
                  tmp.dlina = dlin;
                  tmp.freq = (*iter).freq;
                  pD[iDN].PATH.insert(jt, tmp);
               }else if(dlin < (*jt).dlina){// вставить перед
                  tmp.dlina = dlin;
                  tmp.freq = (*iter).freq;
                  pD[iDN].PATH.insert(jt, tmp);
               }else if(dlin > (*jt).dlina){//
                  while(1){
                     jt++;
                     if(jt == pD[iDN].PATH.end()){
                        tmp.dlina = dlin;
                        tmp.freq = (*iter).freq;
                        pD[iDN].PATH.insert(jt, tmp);
                        break;
                     }else if(dlin == (*jt).dlina){
                        (*jt).freq += (*iter).freq;
                        if((*jt).freq<0) return 3;
                        ++jt;
                        break;
                     }else if(dlin < (*jt).dlina){
                        tmp.dlina = dlin;
                        tmp.freq = (*iter).freq;
                        pD[iDN].PATH.insert(jt, tmp);
                        //                        pD[iDN].PATH.insert(jt, *iter);
                        break;
                     }else{// dlin > (*jt).dlina
                     }
                  }
               }else{// длины равны
                  int fj = (*jt).freq;
                  int fi = (*iter).freq;
                  (*jt).freq += (*iter).freq;
                  if((*jt).freq<0){
                     cout<<" fj="<<fj<<" fi="<<fi;
                     return 3;
                  }
                  ++jt;
               }
            }else{// удалить слишком длинные пути из расчета
               //            qDebug()<<" k="<<k<<" iMN="<<iMN<<" iDN="<<iDN
               //            <<" lnSP="<<sp<<" lnLP="<<pD[iDN].lnLP
               //            <<" past="<<past<<" lnP0="<<lnP0;
               break;// выход из цикла
            }
         }
//      }catch(std::bad_alloc& ba){// нехватка памяти
//         return 2;
//      }
   }// остались пути
   return 0;
}

#endif // NFISHER_H
