#include <QApplication> // отправить события

#include <math.h> // log()
#include <QDebug>
#include <QObject>
#include <QString>
#include <algorithm>
#include "wein.h"

// внешние переменные --------------
extern QApplication *gpApplication;// обработка событий
extern int gStopCalculation; // внешний Останов
extern int gProgressCalc; // прогресс (кол-во столбцов)
int StopCalculation();// останов вычислений

//=================================
// ВХОД:
//   colnum - кол-во аллелей (столбцов)
//   gen - верхняя треугольная матрица генотипов
// ВЫХОД:
//   err - код ошибки (успех =0)
//   pVal - суммарная вероятность
hwein::hwein(int &err, int clm, int *gen, double &pVal){
   err = 0;// успех
   colnum = clm;
   numAlel(gen);// вычисление кол-ва аллелей
   err = initNodes2();// разместить структуры

   // k_step - номер столбца ====================
   for( k_step = colnum; k_step>0; k_step--){// до 1
      for(imN = (*pmN).begin(); imN != (*pmN).end(); ++imN){
         if(StopCalculation()){// внешний останов
            err = 1;
            return;
         }
         if(k_step>1){// вычислять SP и LP
            bool MNfill = true;// остались пути из Mother Node
            if(k_step<colnum){
               // выч.кратчайший путь и отбросить пути >SP
               MNfill = shortPath();
               // выч.длиннейший путь и суммировать пути <LP
               if(MNfill) MNfill = longPath();
            }
            // создать доч.узлы и соединить их с матер.узлом
            if(MNfill) createDN();
         }else{// k_step=1
//            qDebug()<<" last step al="<<(*imN).al[0];
            // суммир.вероятностей на последнем шагу
            P_lastStep();
         }
      }// end MN -------------------
      change_M_D();// обмен указателей MN <-> DN
      gProgressCalc = colnum - k_step + 1;// индикатор прогресса
   }// end k_step =======================
   pVal = Psum;// результат - сумма вероятности
}
//____________________________________________________
// вычислить длиннейший путь и вложить короткие пути в вероятность
bool hwein::longPath(){
   int Nk=(*imN).sumY/2;
   double lnPmax;
   for(int z=0; z<=Nk; z++){
      int d1 = (Nk-z)/k_step;
      int d2 = 2*z/(k_step*(k_step-1));
      double lnP = z*ln2-(Nk-z-k_step*d1)*lnX[d1+1]-
            k_step*lnFact[d1]-
            (z-k_step*(k_step-1)*d2/2.)*lnX[d2+1]-
            k_step*(k_step-1)/2.*lnFact[d2];
      if(z==0) lnPmax = lnP;
      if(lnP>lnPmax) lnPmax = lnP;
   }
   double vComp = lnP0 - lnPmax;
   int ind = less((*imN).lnPath, vComp);
   if(ind<0) return true;// все пути длинные

   double sumLnY=0;
   for(int i=0; i<k_step; i++){
      sumLnY += lnFact[(*imN).al[i]];
   }
   // вклад от суммы всех ребер после материнского узла
   double Pall = lnFact[2*Nk]-lnFact[Nk]-sumLnY;
   vector<double>::const_iterator it=(*imN).lnPath.begin();
//   vector<double>::iterator it=(*imN).lnPath.begin();
   for(; it != (*imN).lnPath.begin()+ind+1; it++){
      Psum += exp(*it + Pall - lnD);
//      (*imN).lnPath.erase(it);
   }
   // выбросить длинные пути
   (*imN).lnPath.erase((*imN).lnPath.begin(),
   (*imN).lnPath.begin()+ind+1);

   bool isPath = true;
   if((*imN).lnPath.size()==0) isPath=false;
   return isPath;
}

//_______________________________________
class ubivorder2{
public: bool operator()(const para2 &p, const para2 &q) const
   {
//      return p.sumAlel <= q.sumAlel;// по возрастанию
      return p.sumAlel >= q.sumAlel;   // по убыванию
   }
};
//_____________________________
// вычислить кратчайший путь и выбросить длинные пути
bool hwein::shortPath(){
   // сортировка по убыванию кол-ва аллелей
   for(int i=0; i<k_step; i++){
      sAlel[i].row = i;
      sAlel[i].sumAlel = (*imN).al[i];
   }
   sort(sAlel, sAlel+k_step, ubivorder2());
   // формирование множеств ==============
   double sumE=0, sumZ=0, sumT=0;
   int numT=0, sumYZ=0;
   bool isCopy = false;
   for(int i=0; i<k_step; i++){
      if(isCopy){
         isCopy = false;
         continue;
      }
      if(sAlel[i].sumAlel%2 == 0){// even
         sumE += lnFact[sAlel[i].sumAlel/2];
      }else{   // odd
         // есть копия?
         if(i != k_step-1 ){// есть копия?
            isCopy = (sAlel[i].sumAlel==sAlel[i+1].sumAlel);
         }
         if(isCopy){
            sumZ += 2*lnFact[sAlel[i].sumAlel];
            sumYZ += 2*sAlel[i].sumAlel;
         }else{
            sumT += lnFact[(sAlel[i].sumAlel-1)/2];
            numT++;
         }
      }
   }
// end формирование множеств ==============
   double lnSP = 0.5*(sumYZ+numT)*ln2 -
               sumE - sumT - 0.5*sumZ;
   double vComp = lnP0 - lnSP;
   int ind = bigger((*imN).lnPath, vComp);
   if(ind<0) return true;// все пути короткие
// выбросить длинные пути
   (*imN).lnPath.erase((*imN).lnPath.begin()+ind,
                       (*imN).lnPath.end());
   bool isPath = true;
   if((*imN).lnPath.size()==0) isPath=false;
   return isPath;
}
//_____________________________
// последний шаг k_step=1
// суммирование вероятностей на последнем шагу
void hwein::P_lastStep(){
   int gz = ((*imN).al[k_step-1])/2;
   double lnLink = -lnFact[gz];
//   qDebug()<<" P_lastStep gz="<<gz;
   int ind = less((*imN).lnPath, lnP0 - lnLink);
   for(int i=0; i<=ind; i++){
      double dist = (*imN).lnPath[i] + lnLink;
//      qDebug()<<" dist="<<dist<<" i="<<i;
      double padd = exp(dist - lnD);
      Psum += padd;
   }
}
//_____________________________
// возврат индекса вектора v[i] > vComp
// возврат -1 когда v[last] <= vComp
int hwein::bigger(vector<double> &cnt, double vComp){
   int iv = less(cnt, vComp);
   if(iv<0) return 0;
   int is = cnt.size();
   if(iv==(is-1)) return -1;
   return iv+1;
}
//_____________________________
// возврат индекса вектора v[i] <= vComp
// возврат -1 когда v[0]>vComp
// возврат -2 когда контейнер пуст
// методом деления отрезка пополам
int hwein::less(vector<double> &cnt, double vComp){
   int ib = cnt.size()-1;
   if(ib<0) return -2;
   int ia = 0, im;
   double tmp = cnt[ib] ;
   if( tmp <= vComp) return ib;
   tmp = cnt[ia];
   if( tmp > vComp) return -1;
   while((ib - ia) > 1){
      im = (ia+ib)/2;
      tmp = cnt[im];
      if(tmp <= vComp){
         ia = im;
      }else{
         ib = im;
      }
   }
   return ia;
}
//_____________________________
// обмен указателей M <=> D
void hwein::change_M_D(){
   list<node2> *pN; //
   pN = pmN;
   pmN = pdN;
   pdN = pN;

  (*pN).clear();
}
//_______________________________________
// создать дочерние узлы и соединить их с материнским узлом
int hwein::createDN(){
   int k = k_step - 2;// индекс последнего аллеля в DN
   // кол-во последнего аллеля, когда он весь в гетерозиготах
   int lbnd = 2*(*imN).al[k_step-1]-(*imN).sumY;
   // нижняя граница последнего генотипа (гомозигота)
   lowbound[k] = MAX(0, (lbnd + 1)/2);
   // верхняя граница последнего генотипа (гомозигота)
   upperbound[k] = (*imN).al[k_step-1]/2;
   // кол-во последнего аллеля
   tAlel[k+1] = (*imN).al[k_step-1];
// ОТЛАДКА -------------
//   qDebug()<<"============================";
//   QString tmp=" ";
//   for(int i=0; i<k_step; i++){
//      tmp += QObject::tr("%1").arg((*imN).al[i]) + ", ";
//   }
//   qDebug()<<" k="<<k<<" lbnd="<<lbnd
//   <<" Mother al="<<tmp <<" lowbound="
//   <<lowbound[k]<<" upperbound="<<upperbound[k]
//   <<" tAlel[k+1]="<<tAlel[k+1];
// ОТЛАДКА -------------
   for_N(k);
   return 0;
}
//_______________________________________
// рекурсивная функция вычисления вектора аллелей Yi в дочернем узле
// n - номер аллеля дочернего узла: n = k_step-2,...,0
//
void hwein::for_N(int n) {
   for (int i = lowbound[n]; i <= upperbound[n]; ++i) {
      tIndex[n] = i;
      if(n > 0){
         if(n==k_step-2){// первый шаг, гомозигота
            tAlel[n] = tAlel[n+1] - 2*i;
         }else{// последующие шаги, гетерозигота
            tAlel[n] = tAlel[n+1] - i;
         }
         int up = MIN(tAlel[n], (*imN).al[n]);
//         qDebug()<<" n="<<n<<" tAlel[n]="<<tAlel[n]<<" up="<<up;
         lowbound[n-1] = 0;// ???
         upperbound[n-1] = up;
         for_N(n - 1);  // рекурсия
      }else{ // n=0, выход из рекурсии
         daughter();// получить дочерний узел
      }
   }
}
//_______________________________________
// создать дочерний узел
// добавить в него длины путей
void hwein::daughter(){
   node2 nod;
   int xg=1;
   if(k_step==2){// 1 слагаемое
      xg=2;// гомозигота
   }
   int sum = 0, num;
   for(int i=0; i<k_step-1; i++){// (k_step-1)  слагаемых
      if(i==0){
         num = (*imN).al[i] - (tAlel[1] - xg*tIndex[i]);
      }else{
         int j=i-1;
         num = (*imN).al[i] - tIndex[j];
      }
//      qDebug()<<" i="<<i<<" alel="<<num;
      if(num < 0) return;  // отрицательное кол-во аллеля
      sum += num;
      nod.al.push_back(num);
   }
// отладка ----------------
//   QString stm="";
//   for(int i=0; i<k_step-1; i++){// (k_step-1)  слагаемых
//      stm += QString(" %1,").arg(nod.al[i]);
//   }
//   qDebug()<<" Dauter allel="<< stm;
// отладка ----------------

//   sum = 2*num - sum;
   nod.sumY = sum;

   // проверить его наличие в списке
   int ind = isDauter(nod);
   if(ind >= 0){// узел есть
      // добавить связь
//      (*imN).lIndex.push_back(ind);
   }else{// узла нет
//      (*imN).lIndex.push_back(indDN++);// добавить связь
      (*pdN).push_back(nod);// добавить узел
      idN = (*pdN).end();
      --idN;// указывает на добавленный узел
   }
   // логарифм длины связи
   double lnLink = lnLen(nod);
//   Past.clear();
   vector<double>::const_iterator it;
//   vector<double>::iterator itD;

   int PathSize = (*idN).lnPath.size();
//   itD = (*idN).lnPath.end();
   int ip=0;
   for(it=(*imN).lnPath.begin(); it!=(*imN).lnPath.end(); ++it){
      double dist = *it + lnLink;
//      Past.push_back(dist);
      (*idN).lnPath.push_back(dist);// добавить в конец
      ip++;
   }
   // сортировка по возрастанию пути
   if(PathSize){
      inplace_merge((*idN).lnPath.begin(),
//      (*idN).lnPath.begin()+ip, (*idN).lnPath.end());
      (*idN).lnPath.begin()+PathSize, (*idN).lnPath.end());
   }
}
//_______________________________________
// логарифм длины связи
double hwein::lnLen(node2 &nod){
   double lnSum=0;
   int isum=0;
   for(int i=0; i<k_step-1; i++){
      int raz = (*imN).al[i]-nod.al[i];
      isum += raz;
      lnSum += lnFact[raz];
   }
   int gz = ((*imN).al[k_step-1]-isum)/2;
   double res = isum*ln2 -(lnFact[gz]+lnSum);
   return res;// отладка
}
//_______________________________________
// есть в списке дочерних узлов
int hwein::isDauter(node2 &node){
   int i=-1;
   for(idN=(*pdN).begin(); idN!=(*pdN).end(); ++idN){
     i++;
     int j;
     for(j=0; j<k_step-1; j++){
        if(node.al[j] != (*idN).al[j]) break;
     }
     if(j==(k_step-1)) return i;  // есть совпадение
   }
   return -1;// нет совпадения
}
//_______________________________________
// деструктор
hwein::~hwein(){
//   delete [] yEven;
//   delete [] yOddZ;
//   delete [] yOddT;
   delete [] sAlel;
   delete [] tIndex;
   delete [] tAlel;
   delete [] alel;
   delete [] gentype;
   delete [] lowbound;
   delete [] upperbound;
   delete [] lnFact;
   delete [] lnX;
}
//_______________________________________
// расчет факториалов, вероятности исходного пути
int hwein::lnfact(int maxFact){
   double logFac = 0;
   lnFact[0] = 0;
   lnX[0]=0;// на всякий случай
   for(int i=1; i<=maxFact; i++){
      double loge = log(i);
      lnX[i] = loge;
      logFac += loge;
      lnFact[i] = logFac;
//      qDebug()<<" i="<<i<<" lnFact[i]="<<lnFact[i];
   }
   ln2 = lnFact[2];
   double znam = 0;
   for(int i=0; i<colnum; i++){
      znam += lnFact[alel[i]];
   }
   lnD = lnFact[2*Ntotal]-lnFact[Ntotal]-znam;
   znam = 0;
   for(int i = 0; i<colnum; i++){
      for(int j = i; j<colnum; j++){
         znam += lnFact[gentype[j+i*colnum]];
      }
   }
//   lnP0 = (Ntotal-Nhomo)*ln2-znam-lnD;
   lnP0 = (Ntotal-Nhomo)*ln2-znam;
//   double cr = EPS;
//   if(lnP0<0){
//      cr = -cr;
//   }
//   lnP0 *= (1+cr);
   lnP0 += EPSWEIN;
   return 0;
}
//_______________________________________
// инициализация
int hwein::initNodes2(){
   node2 nod;
   int sum = 0;
   for(int i=0; i<colnum; ++i){
      nod.al.push_back(alel[i]);
//      if(i<colnum-1){
         sum += alel[i];
//      }
   }
   nod.sumY = sum;
   nod.lnPath.push_back(0);
   mNode.push_back(nod);
   pmN = &mNode;
   pdN = &dNode;
   Psum = 0;
   return 0;
}
//_______________________________________
// вычисление кол-ва аллелей по генотипам
void hwein::numAlel(int *gen){
// заполнение симметричной матрицы генотипов
   gentype = new int [colnum*colnum];
   Ntotal = 0;
   Nhomo = 0;
   for(int i = 0; i<colnum; i++){
      for(int j = 0; j<colnum; j++){
         if(j>=i){// выше диагонали
            int num = gen[j+i*colnum];
            gentype[j+i*colnum] = num;
            Ntotal += num;
            if(j==i) Nhomo += num;
         }else{   // ниже диагонали (симметрия)
            gentype[j+i*colnum] = gen[i+j*colnum];
         }
      }
   }
// заполнение вектора аллелей
   alel = new int [colnum];
   for(int i=0; i<colnum; i++){
      alel[i] = 0;
      for(int j = 0; j<colnum; j++){
         int mnoj = 1;
         if(j==i){
            mnoj = 2;
         }
         alel[i] += mnoj*gentype[j+i*colnum];
      }
//      qDebug()<<" i="<<alel[i];
   }
// сортировка по убыванию кол-ва аллелей
   sAlel = new para2 [colnum];
   for(int i=0; i<colnum; i++){
      sAlel[i].row = i;
      sAlel[i].sumAlel = alel[i];
   }
   sort(sAlel, sAlel+colnum, ubivorder2());

   for(int i=0; i<colnum; i++){
      alel[i] = sAlel[i].sumAlel;
//      qDebug()<<" i="<<i<<" alel="<<alel[i]
//      <<" row="<<sAlel[i].row;
   }
   lnX = new double[2*Ntotal+1];// =ln(x)
   lnFact = new double[2*Ntotal+1];// lnFact[]=ln(x!)
   lnfact(2*Ntotal);
   lowbound = new int [colnum+1];
   upperbound = new int [colnum+1];
   tAlel = new int [colnum+1];
   tIndex = new int [colnum+1];
}
