//==================================
// Сетевой алгоритм равновесия Харди-Вайнберга
//==================================

#ifndef HWEIN_H
#define HWEIN_H

#include <cstdlib>
#include <vector>
#include <list>
using namespace std;

#define EPSWEIN 1e-10 // точность вычислений
#define MIN(X,Y) ((X)<(Y) ? (X) : (Y))
#define MAX(X,Y) ((X)>(Y) ? (X) : (Y))

// структура для сортировки аллелей по убыванию
struct para2{
   int row;    // номер строки
   int sumAlel;// кол-во алелей
};
// структура узла
struct node2{
   int sumY;   // сумма аллелей = sum{al[i]}
// кол-во аллелей в MN = номеру шага k_step
   vector<int> al;   // для DN кол-во аллелей = (k_step-1)
   vector<double> lnPath;// логарифм длины пройденных путей
};

class hwein{
public:
   hwein(int &err, int colnum, int *gen, double &pVal);// конструктор
   ~hwein();   // деструктор

private:
   int colnum;    // кол-во столбцов
   int k_step;    // номер шага
   int *tIndex;   // текущий индекс
   int *tAlel;    // аллели дочернего узла
   para2 *sAlel;  // сортировка строк
   int *gentype;  // симметричная матрица генотипов
   int *alel;     // кол-во аллелей
   int *lowbound;    // нижняя граница аллеля
   int *upperbound;  // верхняя граница аллеля
   double *lnFact;// ln(x!)
   double *lnX;   // ln(x)
   double Psum;   // сумма вероятностей
   double lnP0;   // логарифм исходного пути + lnD
   double lnD;    // множитель расчета вероятности ln(D)
   double ln2;    // log(2)
   int Ntotal;    // сумма генотипов
   int Nhomo;     // кол-во гомозигот
   vector<double> Past;// сумма логарифмов путей
   list<node2> mNode; // mother
   list<node2> dNode; // dauther
   list<node2> *pmN;  // mother
   list<node2> *pdN;  // dauther
   list<node2>::iterator imN;// указатель на mNode
   list<node2>::iterator idN;// указатель на dNode

// функции -----------
   bool longPath();
   bool shortPath();
   int bigger(vector<double> &cont, double vComp);
   int less(vector<double> &cont, double vComp);
   void P_lastStep();// последний шаг
   double lnLen(node2 &nod);
   void change_M_D();
   int isDauter(node2 &nod);
   void daughter();  // создать дочерний узел
   void for_N(int n);   // рекурсивная функция
   int createDN();
   int lnfact(int maxFact);// логарифм факториала
   void numAlel(int *gen);// вычисление кол-ва аллелей по генотипам
   int initNodes2();// разместить структуры
};

#endif // HWEIN_H